import numpy as np
import pytest

from qibojit.backends import MetaBackend


def test_device_setter(backend):
    if backend.platform == "numba":
        device = "/CPU:0"
    else:  # pragma: no cover
        # CI does not have GPUs
        device = "/GPU:0"
    backend.set_device(device)
    assert backend.device == device


def test_thread_setter(backend):
    import numba

    original_threads = numba.get_num_threads()
    backend.set_threads(1)
    assert numba.get_num_threads() == 1


@pytest.mark.parametrize("array_type", [None, "float32", "float64"])
def test_cast(backend, array_type):
    target = np.random.random(10)
    final = backend.to_numpy(backend.cast(target, dtype=array_type))
    backend.assert_allclose(final, target)


@pytest.mark.parametrize("array_type", [None, "float32", "float64"])
@pytest.mark.parametrize("format", ["coo", "csr", "csc", "dia"])
def test_sparse_cast(backend, array_type, format):
    from scipy import sparse

    sptarget = sparse.rand(512, 512, dtype=array_type, format=format)
    assert backend.is_sparse(sptarget)
    final = backend.to_numpy(backend.cast(sptarget))
    target = sptarget.toarray()
    backend.assert_allclose(final, target)
    if backend.platform != "numba":  # pragma: no cover
        sptarget = getattr(backend.sparse, sptarget.__class__.__name__)(sptarget)
        assert backend.is_sparse(sptarget)
        final = backend.to_numpy(backend.cast(sptarget))
        backend.assert_allclose(final, target)


def test_to_numpy(backend):
    x = [0, 1, 2]
    target = backend.to_numpy(backend.cast(x))
    if backend.platform == "numba":
        final = backend.to_numpy(x)
    else:  # pragma: no cover
        final = backend.to_numpy(np.array(x))
    backend.assert_allclose(final, target)


@pytest.mark.parametrize("sparse_type", ["coo", "csr", "csc", "dia"])
def test_backend_expm_sparse(backend, sparse_type):
    from scipy.linalg import expm
    from scipy.sparse import rand

    m = rand(16, 16, format=sparse_type)
    target = expm(m.toarray())
    result = backend.to_numpy(backend.calculate_matrix_exp(1j, backend.cast(m)))
    backend.assert_allclose(target, result, atol=1e-10)


@pytest.mark.parametrize("sparse_type", [None, "coo", "csr", "csc", "dia"])
def test_backend_eigh(backend, sparse_type):
    if sparse_type is None:
        m = np.random.random((16, 16))
        eigvals1, eigvecs1 = backend.calculate_eigenvectors(backend.cast(m))
        eigvals2, eigvecs2 = np.linalg.eigh(m)
    else:
        from scipy.sparse import rand

        m = rand(16, 16, format=sparse_type)
        m = m + m.T
        eigvals1, eigvecs1 = backend.calculate_eigenvectors(backend.cast(m), k=16)
        eigvals2, eigvecs2 = backend.calculate_eigenvectors(backend.cast(m.toarray()))
    backend.assert_allclose(eigvals1, eigvals2, atol=1e-10)
    eigvecs1 = backend.to_numpy(eigvecs1)
    eigvecs2 = backend.to_numpy(eigvecs2)
    backend.assert_allclose(np.abs(eigvecs1), np.abs(eigvecs2), atol=1e-10)


@pytest.mark.parametrize("sparse_type", [None, "coo", "csr", "csc", "dia"])
def test_backend_eigvalsh(backend, sparse_type):
    if sparse_type is None:
        m = np.random.random((16, 16))
        target = np.linalg.eigvalsh(m)
        result = backend.calculate_eigenvalues(backend.cast(m))
    else:
        from scipy.sparse import rand

        m = rand(16, 16, format=sparse_type)
        m = m + m.T
        result = backend.calculate_eigenvalues(backend.cast(m), k=16)
        target, _ = backend.calculate_eigenvectors(backend.cast(m.toarray()))
    backend.assert_allclose(target, result, atol=1e-10)


@pytest.mark.parametrize("sparse_type", ["coo", "csr", "csc", "dia"])
@pytest.mark.parametrize("k", [6, 8])
def test_backend_eigh_sparse(backend, sparse_type, k):
    from qibo import hamiltonians
    from scipy import sparse
    from scipy.sparse.linalg import eigsh

    ham = hamiltonians.TFIM(6, h=1.0, backend=backend)
    m = getattr(sparse, f"{sparse_type}_matrix")(backend.to_numpy(ham.matrix))
    eigvals1, eigvecs1 = backend.calculate_eigenvectors(backend.cast(m), k)
    eigvals2, eigvecs2 = eigsh(m, k, which="SA")
    eigvals1 = backend.to_numpy(eigvals1)
    eigvals2 = backend.to_numpy(eigvals2)
    backend.assert_allclose(sorted(eigvals1), sorted(eigvals2))


def test_metabackend_list_available():
    available_backends = dict(zip(("numba", "cupy", "cuquantum"), (True, False, False)))
    assert MetaBackend().list_available() == available_backends



"""Set of custom CuPy operations for the Clifford backend."""

import cupy as cp  # pylint: disable=E0401
import numpy
from qibo.backends._clifford_operations import _dim, _get_rxz, _packed_size, _pad_size
from scipy import sparse

np = cp

GRIDDIM, BLOCKDIM = 1024, 128
GRIDDIM_2D = (1024, 1024)


apply_one_qubit_kernel = """
extern "C"
__global__ void apply_{}(unsigned char* symplectic_matrix, const int q, const int qz, const int nrows, const int ncolumns) {{
    _apply_{}(symplectic_matrix, q, qz, nrows, ncolumns);
}}
"""

apply_two_qubits_kernel = """
extern "C"
__global__ void apply_{}(unsigned char* symplectic_matrix, const int control_q, const int target_q, const int cqz, const int tqz, const int nrows, const int ncolumns) {{
    _apply_{}(symplectic_matrix, control_q, target_q, cqz, tqz, nrows, ncolumns);
}}
"""


def one_qubit_kernel_launcher(kernel, symplectic_matrix, q, nqubits):
    qz = nqubits + q
    ncolumns = _dim(nqubits)
    nrows = _packed_size(ncolumns)
    return kernel((GRIDDIM,), (BLOCKDIM,), (symplectic_matrix, q, qz, nrows, ncolumns))


def two_qubits_kernel_launcher(kernel, symplectic_matrix, control_q, target_q, nqubits):
    cqz = nqubits + control_q
    tqz = nqubits + target_q
    ncolumns = _dim(nqubits)
    nrows = _packed_size(ncolumns)
    return kernel(
        (GRIDDIM,),
        (BLOCKDIM,),
        (symplectic_matrix, control_q, target_q, cqz, tqz, nrows, ncolumns),
    )


apply_H = """
__device__ void _apply_H(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (symplectic_matrix[row_idx + q] & symplectic_matrix[row_idx + qz]);
        const unsigned char tmp = symplectic_matrix[row_idx + q];
        symplectic_matrix[row_idx + q] = symplectic_matrix[row_idx + qz];
        symplectic_matrix[row_idx + qz] = tmp;
    };
}
""" + apply_one_qubit_kernel.format(
    "H", "H"
)

apply_H = cp.RawKernel(apply_H, "apply_H", options=("--std=c++11",))


def H(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_H, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_CNOT = """
__device__ void _apply_CNOT(unsigned char* symplectic_matrix, const int& control_q, const int& target_q, const int& cqz, const int& tqz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (
            symplectic_matrix[row_idx + control_q] & symplectic_matrix[row_idx + tqz]
        ) & (symplectic_matrix[row_idx + target_q] ^ ~symplectic_matrix[row_idx + cqz]);
        symplectic_matrix[row_idx + target_q] = (
            symplectic_matrix[row_idx + target_q] ^ symplectic_matrix[row_idx + control_q]
        );
        symplectic_matrix[row_idx + cqz] = (
            symplectic_matrix[row_idx + cqz] ^ symplectic_matrix[row_idx + tqz]
        );
    };
}
""" + apply_two_qubits_kernel.format(
    "CNOT", "CNOT"
)

apply_CNOT = cp.RawKernel(apply_CNOT, "apply_CNOT", options=("--std=c++11",))


def CNOT(symplectic_matrix, control_q, target_q, nqubits):
    two_qubits_kernel_launcher(
        apply_CNOT, symplectic_matrix, control_q, target_q, nqubits
    )
    return symplectic_matrix


apply_CZ = """
__device__ void _apply_CZ(unsigned char* symplectic_matrix, const int& control_q, const int& target_q, const int& cqz, const int& tqz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = (
            symplectic_matrix[row_idx + last]
            ^ (symplectic_matrix[row_idx + target_q] & symplectic_matrix[row_idx + tqz])
            ^ (
                symplectic_matrix[row_idx + control_q]
                & symplectic_matrix[row_idx + target_q]
                & (symplectic_matrix[row_idx + tqz] ^ ~symplectic_matrix[row_idx + cqz])
            )
            ^ (
                symplectic_matrix[row_idx + target_q]
                & (symplectic_matrix[row_idx + tqz] ^ symplectic_matrix[row_idx + control_q])
            )
        );
        const unsigned char z_control_q = symplectic_matrix[row_idx + target_q] ^ symplectic_matrix[row_idx + cqz];
        const unsigned char z_target_q = symplectic_matrix[row_idx + tqz] ^ symplectic_matrix[row_idx + control_q];
        symplectic_matrix[row_idx + cqz] = z_control_q;
        symplectic_matrix[row_idx + tqz] = z_target_q;
    };
}
""" + apply_two_qubits_kernel.format(
    "CZ", "CZ"
)

apply_CZ = cp.RawKernel(apply_CZ, "apply_CZ", options=("--std=c++11",))


def CZ(symplectic_matrix, control_q, target_q, nqubits):
    two_qubits_kernel_launcher(
        apply_CZ, symplectic_matrix, control_q, target_q, nqubits
    )
    return symplectic_matrix


apply_S = """
__device__ void _apply_S(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (
            symplectic_matrix[row_idx + q] & symplectic_matrix[row_idx + qz]
        );
        symplectic_matrix[row_idx + qz] = symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q];
    };
}
""" + apply_one_qubit_kernel.format(
    "S", "S"
)

apply_S = cp.RawKernel(apply_S, "apply_S", options=("--std=c++11",))


def S(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_S, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_Z = """
__device__ void _apply_Z(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (
            (symplectic_matrix[row_idx + q] & symplectic_matrix[row_idx + qz])
            ^ symplectic_matrix[row_idx + q]
            & (symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q])
        );
    };
}
""" + apply_one_qubit_kernel.format(
    "Z", "Z"
)

apply_Z = cp.RawKernel(apply_Z, "apply_Z", options=("--std=c++11",))


def Z(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_Z, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_X = """
__device__ void _apply_X(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = (
            symplectic_matrix[row_idx + last]
            ^ (
                symplectic_matrix[row_idx + qz]
                & (symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q])
            )
            ^ (symplectic_matrix[row_idx + qz] & symplectic_matrix[row_idx + q])
        );
    };
}
""" + apply_one_qubit_kernel.format(
    "X", "X"
)

apply_X = cp.RawKernel(apply_X, "apply_X", options=("--std=c++11",))


def X(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_X, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_Y = """
__device__ void _apply_Y(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = (
            symplectic_matrix[row_idx + last]
            ^ (
                symplectic_matrix[row_idx + qz]
                & (symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q])
            )
            ^ (
                symplectic_matrix[row_idx + q]
                & (symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q])
            )
        );
    };
}
""" + apply_one_qubit_kernel.format(
    "Y", "Y"
)

apply_Y = cp.RawKernel(apply_Y, "apply_Y", options=("--std=c++11",))


def Y(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_Y, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_SX = """
__device__ void _apply_SX(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (
            symplectic_matrix[row_idx + qz]
            & (symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q])
        );
        symplectic_matrix[row_idx + q] = symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q];
    };
}
""" + apply_one_qubit_kernel.format(
    "SX", "SX"
)

apply_SX = cp.RawKernel(apply_SX, "apply_SX", options=("--std=c++11",))


def SX(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_SX, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_SDG = """
__device__ void _apply_SDG(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (
            symplectic_matrix[row_idx + q]
            & (symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q])
        );
        symplectic_matrix[row_idx + qz] = symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q];
    };
}
""" + apply_one_qubit_kernel.format(
    "SDG", "SDG"
)

apply_SDG = cp.RawKernel(apply_SDG, "apply_SDG", options=("--std=c++11",))


def SDG(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_SDG, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_SXDG = """
__device__ void _apply_SXDG(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (
            symplectic_matrix[row_idx + qz] & symplectic_matrix[row_idx + q]
        );
        symplectic_matrix[row_idx + q] = symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q];
    };
}
""" + apply_one_qubit_kernel.format(
    "SXDG", "SXDG"
)

apply_SXDG = cp.RawKernel(apply_SXDG, "apply_SXDG", options=("--std=c++11",))


def SXDG(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_SXDG, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_RY_pi = """
__device__ void _apply_RY_pi(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (
            symplectic_matrix[row_idx + q]
            & (symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q])
        );
        const unsigned char zq = symplectic_matrix[row_idx + qz];
        symplectic_matrix[row_idx + qz] = symplectic_matrix[row_idx + q];
        symplectic_matrix[row_idx + q] = zq;
    };
}
""" + apply_one_qubit_kernel.format(
    "RY_pi", "RY_pi"
)

apply_RY_pi = cp.RawKernel(apply_RY_pi, "apply_RY_pi", options=("--std=c++11",))


def RY_pi(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_RY_pi, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_RY_3pi_2 = """
__device__ void _apply_RY_3pi_2(unsigned char* symplectic_matrix, const int& q, const int& qz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = symplectic_matrix[row_idx + last] ^ (
            symplectic_matrix[row_idx + qz]
            & (symplectic_matrix[row_idx + qz] ^ symplectic_matrix[row_idx + q])
        );
        const unsigned char zq = symplectic_matrix[row_idx + qz];
        symplectic_matrix[row_idx + qz] = symplectic_matrix[row_idx + q];
        symplectic_matrix[row_idx + q] = zq;
    };
}
""" + apply_one_qubit_kernel.format(
    "RY_3pi_2", "RY_3pi_2"
)

apply_RY_3pi_2 = cp.RawKernel(
    apply_RY_3pi_2, "apply_RY_3pi_2", options=("--std=c++11",)
)


def RY_3pi_2(symplectic_matrix, q, nqubits):
    one_qubit_kernel_launcher(apply_RY_3pi_2, symplectic_matrix, q, nqubits)
    return symplectic_matrix


apply_SWAP = """
__device__ void _apply_SWAP(unsigned char* symplectic_matrix, const int& control_q, const int& target_q, const int& cqz, const int& tqz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = (
            symplectic_matrix[row_idx + last]
            ^ (
                symplectic_matrix[row_idx + control_q]
                & symplectic_matrix[row_idx + tqz]
                & (symplectic_matrix[row_idx + target_q] ^ ~symplectic_matrix[row_idx + cqz])
            )
            ^ (
                (symplectic_matrix[row_idx + target_q] ^ symplectic_matrix[row_idx + control_q])
                & (symplectic_matrix[row_idx + tqz] ^ symplectic_matrix[row_idx + cqz])
                & (symplectic_matrix[row_idx + tqz] ^ ~symplectic_matrix[row_idx + control_q])
            )
            ^ (
                symplectic_matrix[row_idx + target_q]
                & symplectic_matrix[row_idx + cqz]
                & (
                    symplectic_matrix[row_idx + control_q]
                    ^ symplectic_matrix[row_idx + target_q]
                    ^ symplectic_matrix[row_idx + cqz]
                    ^ ~symplectic_matrix[row_idx + tqz]
                )
            )
        );
        const unsigned char x_cq = symplectic_matrix[row_idx + control_q];
        const unsigned char x_tq = symplectic_matrix[row_idx + target_q];
        const unsigned char z_cq = symplectic_matrix[row_idx + cqz];
        const unsigned char z_tq = symplectic_matrix[row_idx + tqz];
        symplectic_matrix[row_idx + control_q] = x_tq;
        symplectic_matrix[row_idx + target_q] = x_cq;
        symplectic_matrix[row_idx + cqz] = z_tq;
        symplectic_matrix[row_idx + tqz] = z_cq;
    };
}
""" + apply_two_qubits_kernel.format(
    "SWAP", "SWAP"
)

apply_SWAP = cp.RawKernel(apply_SWAP, "apply_SWAP", options=("--std=c++11",))


def SWAP(symplectic_matrix, control_q, target_q, nqubits):
    two_qubits_kernel_launcher(
        apply_SWAP, symplectic_matrix, control_q, target_q, nqubits
    )
    return symplectic_matrix


apply_iSWAP = """
__device__ void _apply_iSWAP(unsigned char* symplectic_matrix, const int& control_q, const int& target_q, const int& cqz, const int& tqz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = (
            symplectic_matrix[row_idx + last]
            ^ (symplectic_matrix[row_idx + target_q] & symplectic_matrix[row_idx + tqz])
            ^ (symplectic_matrix[row_idx + control_q] & symplectic_matrix[row_idx + cqz])
            ^ (
                symplectic_matrix[row_idx + control_q]
                & (symplectic_matrix[row_idx + cqz] ^ symplectic_matrix[row_idx + control_q])
            )
            ^ (
                (symplectic_matrix[row_idx + cqz] ^ symplectic_matrix[row_idx + control_q])
                & (symplectic_matrix[row_idx + tqz] ^ symplectic_matrix[row_idx + target_q])
                & (symplectic_matrix[row_idx + target_q] ^ ~symplectic_matrix[row_idx + control_q])
            )
            ^ (
                (
                    symplectic_matrix[row_idx + target_q]
                    ^ symplectic_matrix[row_idx + cqz]
                    ^ symplectic_matrix[row_idx + control_q]
                )
                & (
                    symplectic_matrix[row_idx + target_q]
                    ^ symplectic_matrix[row_idx + tqz]
                    ^ symplectic_matrix[row_idx + control_q]
                )
                & (
                    symplectic_matrix[row_idx + target_q]
                    ^ symplectic_matrix[row_idx + tqz]
                    ^ symplectic_matrix[row_idx + control_q]
                    ^ ~symplectic_matrix[row_idx + cqz]
                )
            )
            ^ (
                symplectic_matrix[row_idx + control_q]
                & (
                    symplectic_matrix[row_idx + target_q]
                    ^ symplectic_matrix[row_idx + control_q]
                    ^ symplectic_matrix[row_idx + cqz]
                )
            )
        );
        const unsigned char z_control_q = (
            symplectic_matrix[row_idx + target_q]
            ^ symplectic_matrix[row_idx + tqz]
            ^ symplectic_matrix[row_idx + control_q]
        );
        const unsigned char z_target_q = (
            symplectic_matrix[row_idx + target_q]
            ^ symplectic_matrix[row_idx + cqz]
            ^ symplectic_matrix[row_idx + control_q]
        );
        symplectic_matrix[row_idx + cqz] = z_control_q;
        symplectic_matrix[row_idx + tqz] = z_target_q;
        const unsigned char tmp = symplectic_matrix[row_idx + control_q];
        symplectic_matrix[row_idx + control_q] = symplectic_matrix[row_idx + target_q];
        symplectic_matrix[row_idx + target_q] = tmp;
    };
}
""" + apply_two_qubits_kernel.format(
    "iSWAP", "iSWAP"
)

apply_iSWAP = cp.RawKernel(apply_iSWAP, "apply_iSWAP", options=("--std=c++11",))


def iSWAP(symplectic_matrix, control_q, target_q, nqubits):
    two_qubits_kernel_launcher(
        apply_iSWAP, symplectic_matrix, control_q, target_q, nqubits
    )
    return symplectic_matrix


apply_CY = """
__device__ void _apply_CY(unsigned char* symplectic_matrix, const int& control_q, const int& target_q, const int& cqz, const int& tqz, const int& nrows, const int& ncolumns) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int ntid = gridDim.x * blockDim.x;
    const int last = ncolumns - 1;
    for(int i = tid; i < nrows; i += ntid) {
        unsigned int row_idx = i * ncolumns;
        symplectic_matrix[row_idx + last] = (
            symplectic_matrix[row_idx + last]
            ^ (
                symplectic_matrix[row_idx + target_q]
                & (symplectic_matrix[row_idx + tqz] ^ symplectic_matrix[row_idx + target_q])
            )
            ^ (
                symplectic_matrix[row_idx + control_q]
                & (symplectic_matrix[row_idx + target_q] ^ symplectic_matrix[row_idx + tqz])
                & (symplectic_matrix[row_idx + cqz] ^ ~symplectic_matrix[row_idx + target_q])
            )
            ^ (
                (symplectic_matrix[row_idx + target_q] ^ symplectic_matrix[row_idx + control_q])
                & (symplectic_matrix[row_idx + tqz] ^ symplectic_matrix[row_idx + target_q])
            )
        );
        const unsigned char x_target_q = symplectic_matrix[row_idx + control_q] ^ symplectic_matrix[row_idx + target_q];
        const unsigned char z_control_q = (
            symplectic_matrix[row_idx + cqz]
            ^ symplectic_matrix[row_idx + tqz]
            ^ symplectic_matrix[row_idx + target_q]
        );
        const unsigned char z_target_q = symplectic_matrix[row_idx + tqz] ^ symplectic_matrix[row_idx + control_q];
        symplectic_matrix[row_idx + target_q] = x_target_q;
        symplectic_matrix[row_idx + cqz] = z_control_q;
        symplectic_matrix[row_idx + tqz] = z_target_q;
    };
}
""" + apply_two_qubits_kernel.format(
    "CY", "CY"
)

apply_CY = cp.RawKernel(apply_CY, "apply_CY", options=("--std=c++11",))


def CY(symplectic_matrix, control_q, target_q, nqubits):
    two_qubits_kernel_launcher(
        apply_CY, symplectic_matrix, control_q, target_q, nqubits
    )
    return symplectic_matrix


_apply_rowsum = """
__device__ void _apply_rowsum(unsigned char* symplectic_matrix, const long* h, const long* i, const int& nqubits, const bool& determined, const int& nrows, long* g_exp, const int& dim) {
    unsigned int tid_x = blockIdx.x * blockDim.x + threadIdx.x;
    unsigned int bid_y = blockIdx.y;
    unsigned int ntid_x = gridDim.x * blockDim.x;
    unsigned int nbid_y = gridDim.y;
    const int last = dim - 1;
    __shared__ int exp;
    for(int j = bid_y; j < nrows; j += nbid_y) {
        unsigned int row_i = i[j] * dim;
        unsigned int row_h = h[j] * dim;
        for(int k = tid_x; k < nqubits; k += ntid_x) {
            unsigned int kz = nqubits + k;
            exp = (
                2 * (symplectic_matrix[row_i + k] * symplectic_matrix[row_h + k] * (symplectic_matrix[row_h + kz] - symplectic_matrix[row_i + kz]) +
                symplectic_matrix[row_i + kz] * symplectic_matrix[row_h + kz] * (symplectic_matrix[row_i + k] - symplectic_matrix[row_h + k]))
                - symplectic_matrix[row_i + k] * symplectic_matrix[row_h + kz]
                + symplectic_matrix[row_h + k] * symplectic_matrix[row_i + kz]
            );
        }
        if (threadIdx.x == 0 && tid_x < nqubits) {
            g_exp[j] += exp;
        }
        __syncthreads();
        if (threadIdx.x == 0 && blockIdx.x == 0) {
            symplectic_matrix[row_h + last] = (
                2 * symplectic_matrix[row_h + last] + 2 * symplectic_matrix[row_i + last] + g_exp[j]
            ) % 4 != 0;
        }
        for(int k = tid_x; k < nqubits; k += ntid_x) {
            unsigned int kz = nqubits + k;
            unsigned char xi_xh = (
                symplectic_matrix[row_i + k] ^ symplectic_matrix[row_h + k]
            );
            unsigned char zi_zh = (
                symplectic_matrix[row_i + kz] ^ symplectic_matrix[row_h + kz]
            );
            if (determined) {
                symplectic_matrix[row_h + k] ^= xi_xh;
                symplectic_matrix[row_h + kz] ^= zi_zh;
            } else {
                symplectic_matrix[row_h + k] = xi_xh;
                symplectic_matrix[row_h + kz] = zi_zh;
            }
        }
    }
}
"""

apply_rowsum = f"""
{_apply_rowsum}
extern "C"
__global__ void apply_rowsum(unsigned char* symplectic_matrix, const long* h, const long* i, const int nqubits, const bool determined, const int nrows, long* g_exp, const int dim) {{
    _apply_rowsum(symplectic_matrix, h, i, nqubits, determined, nrows, g_exp, dim);
}}
"""

apply_rowsum = cp.RawKernel(apply_rowsum, "apply_rowsum", options=("--std=c++11",))


def _rowsum(symplectic_matrix, h, i, nqubits, determined=False):
    nrows = len(h)
    exp = cp.zeros(len(h), dtype=int)
    packed_nqubits = _packed_size(nqubits)
    row_dim = _dim(packed_nqubits)
    apply_rowsum(
        GRIDDIM_2D,
        (BLOCKDIM,),
        (symplectic_matrix, h, i, packed_nqubits, determined, nrows, exp, row_dim),
    )
    return symplectic_matrix


def _random_outcome(state, p, q, nqubits):
    p = p[0] + nqubits
    tmp = state[p, q].copy()
    state[p, q] = 0
    h = state[:-1, q].nonzero()[0]
    state[p, q] = tmp
    if h.shape[0] > 0:
        dim = state.shape[1]
        state = _pack_for_measurements(state, nqubits)
        dim = state.shape[1]
        state = _rowsum(
            state.ravel(),
            h,
            p.astype(cp.uint) * cp.ones(h.shape[0], dtype=np.uint),
            _packed_size(nqubits),
            False,
        )
        state = _unpack_for_measurements(state.reshape(-1, dim), nqubits)
    state[p - nqubits, :] = state[p, :]
    outcome = cp.random.randint(2, size=None, dtype=cp.uint)
    state[p, :] = 0
    state[p, -1] = outcome.astype(cp.uint8)
    state[p, nqubits + q] = 1
    return state, outcome


def _determined_outcome(state, q, nqubits):
    state[-1, :] = 0
    idx = (state[:nqubits, q].nonzero()[0] + nqubits).astype(np.uint)
    state = _pack_for_measurements(state, nqubits)
    dim = state.shape[1]
    state = _rowsum(
        state.ravel(),
        (2 * nqubits * cp.ones(idx.shape, dtype=np.uint)).astype(np.uint),
        idx.astype(np.uint),
        _packed_size(nqubits),
        True,
    )
    state = _unpack_for_measurements(state.reshape(-1, dim), nqubits)
    return state, state[-1, -1]


def _packbits(array, axis):
    # cupy.packbits doesn't support axis yet
    return cp.array(numpy.packbits(array.get(), axis=axis), dtype=cp.uint8)


def _unpackbits(array, axis):
    return cp.array(numpy.unpackbits(array.get(), axis=axis), dtype=cp.uint8)


def _pack_for_measurements(state, nqubits):
    r, x, z = _get_rxz(state, nqubits)
    x = _packbits(x, axis=1)
    z = _packbits(z, axis=1)
    return np.hstack((x, z, r[:, None]))


def _unpack_for_measurements(state, nqubits):
    xz = _unpackbits(state[:, :-1], axis=1)
    padding_size = _pad_size(nqubits)
    x, z = xz[:, :nqubits], xz[:, nqubits + padding_size : -padding_size]
    return np.hstack((x, z, state[:, -1][:, None]))


def _init_state_for_measurements(state, nqubits, collapse):
    dim = _dim(nqubits)
    if collapse:
        return _unpackbits(state[None, :], axis=0)[:dim]
    else:
        return state.copy()


def cast(x, dtype=None, copy=False):
    if dtype is None:
        dtype = "complex128"

    if cp.sparse.issparse(x):
        if dtype != x.dtype:
            return x.astype(dtype)
        return x

    if sparse.issparse(x):
        cls = getattr(cp.sparse, x.__class__.__name__)
        return cls(x, dtype=dtype)

    if isinstance(x, cp.ndarray) and copy:
        return cp.copy(cp.asarray(x, dtype=dtype))

    return cp.asarray(x, dtype=dtype)


def _clifford_pre_execution_reshape(state):
    return _packbits(state, axis=0).ravel()


def _clifford_post_execution_reshape(state, nqubits):
    dim = _dim(nqubits)
    return _unpackbits(state.reshape(-1, dim), axis=0)[:dim]


def identity_density_matrix(nqubits, normalize: bool = True):
    n = 1 << nqubits
    state = cp.eye(n, dtype="complex128")
    cp.cuda.stream.get_current_stream().synchronize()
    if normalize:
        state /= 2**nqubits
    return state.reshape((n, n))



import numpy as np
from qibo.backends.numpy import NumpyBackend
from qibo.config import log
from qibo.gates.abstract import ParametrizedGate
from qibo.gates.channels import ReadoutErrorChannel
from qibo.gates.special import FusedGate

from qibojit.backends.matrices import CustomMatrices

GATE_OPS = {
    "X": "apply_x",
    "CNOT": "apply_x",
    "TOFFOLI": "apply_x",
    "Y": "apply_y",
    "Z": "apply_z",
    "CY": "apply_y",
    "CZ": "apply_z",
    "U1": "apply_z_pow",
    "CU1": "apply_z_pow",
    "SWAP": "apply_swap",
    "fSim": "apply_fsim",
    "GeneralizedfSim": "apply_fsim",
}


class NumbaBackend(NumpyBackend):
    def __init__(self):
        super().__init__()
        import sys

        import psutil
        from numba import __version__ as numba_version

        from qibojit import __version__ as qibojit_version
        from qibojit.custom_operators import gates, ops

        self.name = "qibojit"
        self.platform = "numba"
        self.versions.update(
            {
                "qibojit": qibojit_version,
                "numba": numba_version,
            }
        )
        self.numeric_types = (
            int,
            float,
            complex,
            np.int32,
            np.int64,
            np.float32,
            np.float64,
            np.complex64,
            np.complex128,
        )
        self.tensor_types = (np.ndarray,)
        self.device = "/CPU:0"
        self.custom_matrices = CustomMatrices(self.dtype)
        self.gates = gates
        self.ops = ops
        self.measure_frequencies_op = ops.measure_frequencies
        self.multi_qubit_kernels = {
            3: self.gates.apply_three_qubit_gate_kernel,
            4: self.gates.apply_four_qubit_gate_kernel,
            5: self.gates.apply_five_qubit_gate_kernel,
        }
        if sys.platform == "darwin":  # pragma: no cover
            self.set_threads(psutil.cpu_count(logical=False))
        else:
            self.set_threads(len(psutil.Process().cpu_affinity()))

    def set_precision(self, precision):
        if precision != self.precision:
            super().set_precision(precision)
            if self.custom_matrices:
                self.custom_matrices = CustomMatrices(self.dtype)

    def set_threads(self, nthreads):
        import numba

        numba.set_num_threads(nthreads)
        self.nthreads = nthreads

    # def cast(self, x, dtype=None, copy=False): Inherited from ``NumpyBackend``

    # def to_numpy(self, x): Inherited from ``NumpyBackend``

    def zero_state(self, nqubits):
        size = 2**nqubits
        state = np.empty((size,), dtype=self.dtype)
        return self.ops.initial_state_vector(state)

    def zero_density_matrix(self, nqubits):
        size = 2**nqubits
        state = np.empty((size, size), dtype=self.dtype)
        return self.ops.initial_density_matrix(state)

    # def plus_state(self, nqubits): Inherited from ``NumpyBackend``

    # def plus_density_matrix(self, nqubits): Inherited from ``NumpyBackend``

    # def matrix_special(self, gate): Inherited from ``NumpyBackend``

    # def control_matrix(self, gate): Inherited from ``NumpyBackend``

    def one_qubit_base(self, state, nqubits, target, kernel, gate, qubits):
        ncontrols = len(qubits) - 1 if qubits is not None else 0
        m = nqubits - target - 1
        nstates = 1 << (nqubits - ncontrols - 1)
        if ncontrols:
            kernel = getattr(self.gates, "multicontrol_{}_kernel".format(kernel))
            return kernel(state, gate, qubits, nstates, m)
        kernel = getattr(self.gates, "{}_kernel".format(kernel))
        return kernel(state, gate, nstates, m)

    def two_qubit_base(self, state, nqubits, target1, target2, kernel, gate, qubits):
        ncontrols = len(qubits) - 2 if qubits is not None else 0
        if target1 > target2:
            swap_targets = True
            m1 = nqubits - target1 - 1
            m2 = nqubits - target2 - 1
        else:
            swap_targets = False
            m1 = nqubits - target2 - 1
            m2 = nqubits - target1 - 1
        nstates = 1 << (nqubits - 2 - ncontrols)
        if ncontrols:
            kernel = getattr(self.gates, "multicontrol_{}_kernel".format(kernel))
            return kernel(state, gate, qubits, nstates, m1, m2, swap_targets)
        kernel = getattr(self.gates, "{}_kernel".format(kernel))
        return kernel(state, gate, nstates, m1, m2, swap_targets)

    def multi_qubit_base(self, state, nqubits, targets, gate, qubits):
        if qubits is None:
            qubits = np.array(sorted(nqubits - q - 1 for q in targets), dtype="int32")
        nstates = 1 << (nqubits - len(qubits))
        targets = np.array(
            [1 << (nqubits - t - 1) for t in targets[::-1]], dtype="int64"
        )
        if len(targets) > 5:
            kernel = self.gates.apply_multi_qubit_gate_kernel
        else:
            kernel = self.multi_qubit_kernels.get(len(targets))
        return kernel(state, gate, qubits, nstates, targets)

    @staticmethod
    def _create_qubits_tensor(gate, nqubits):
        # TODO: Treat density matrices
        qubits = [nqubits - q - 1 for q in gate.control_qubits]
        qubits.extend(nqubits - q - 1 for q in gate.target_qubits)
        return np.array(sorted(qubits), dtype="int32")

    def _as_custom_matrix(self, gate):
        name = gate.__class__.__name__
        if isinstance(gate, ParametrizedGate):
            return getattr(self.custom_matrices, name)(*gate.parameters)
        elif isinstance(gate, FusedGate):  # pragma: no cover
            # fusion is tested in qibo tests
            return self.matrix_fused(gate)
        else:
            matrix = getattr(self.custom_matrices, name)
            return matrix(2 ** len(gate.target_qubits)) if callable(matrix) else matrix

    def apply_gate(self, gate, state, nqubits):
        matrix = self._as_custom_matrix(gate)
        qubits = self._create_qubits_tensor(gate, nqubits)
        targets = gate.target_qubits
        state = self.cast(state)
        if len(targets) == 1:
            op = GATE_OPS.get(gate.__class__.__name__, "apply_gate")
            return self.one_qubit_base(state, nqubits, *targets, op, matrix, qubits)
        elif len(targets) == 2:
            op = GATE_OPS.get(gate.__class__.__name__, "apply_two_qubit_gate")
            return self.two_qubit_base(state, nqubits, *targets, op, matrix, qubits)
        else:
            return self.multi_qubit_base(state, nqubits, targets, matrix, qubits)

    def apply_gate_density_matrix(self, gate, state, nqubits, inverse=False):
        name = gate.__class__.__name__
        if name in ["Y", "CY"]:
            return self._apply_ygate_density_matrix(gate, state, nqubits)
        if inverse:
            # used to reset the state when applying channels
            # see :meth:`qibojit.backend.NumpyBackend.apply_channel_density_matrix` below
            matrix = np.linalg.inv(gate.matrix(self))
            matrix = self.cast(matrix)
        else:
            matrix = self._as_custom_matrix(gate)
        qubits = self._create_qubits_tensor(gate, nqubits)
        qubits_dm = qubits + nqubits
        targets = gate.target_qubits
        targets_dm = tuple(q + nqubits for q in targets)

        state = self.cast(state)
        shape = state.shape
        if len(targets) == 1:
            op = GATE_OPS.get(name, "apply_gate")
            state = self.one_qubit_base(
                state.ravel(), 2 * nqubits, *targets, op, matrix, qubits_dm
            )
            state = self.one_qubit_base(
                state, 2 * nqubits, *targets_dm, op, np.conj(matrix), qubits
            )
        elif len(targets) == 2:
            op = GATE_OPS.get(name, "apply_two_qubit_gate")
            state = self.two_qubit_base(
                state.ravel(), 2 * nqubits, *targets, op, matrix, qubits_dm
            )
            state = self.two_qubit_base(
                state, 2 * nqubits, *targets_dm, op, np.conj(matrix), qubits
            )
        else:
            state = self.multi_qubit_base(
                state.ravel(), 2 * nqubits, targets, matrix, qubits_dm
            )
            state = self.multi_qubit_base(
                state, 2 * nqubits, targets_dm, np.conj(matrix), qubits
            )
        return np.reshape(state, shape)

    def _apply_ygate_density_matrix(self, gate, state, nqubits):
        matrix = self._as_custom_matrix(gate)
        qubits = self._create_qubits_tensor(gate, nqubits)
        qubits_dm = qubits + nqubits
        targets = gate.target_qubits
        targets_dm = tuple(q + nqubits for q in targets)
        state = self.cast(state)
        shape = state.shape
        state = self.one_qubit_base(
            state.ravel(), 2 * nqubits, *targets, "apply_y", matrix, qubits_dm
        )
        # force using ``apply_gate`` kernel so that conjugate is properly applied
        state = self.one_qubit_base(
            state, 2 * nqubits, *targets_dm, "apply_gate", np.conj(matrix), qubits
        )
        return np.reshape(state, shape)

    # def apply_channel(self, gate): Inherited from ``NumpyBackend``

    def apply_channel_density_matrix(self, channel, state, nqubits):
        state = self.cast(state)
        if not channel._all_unitary_operators:
            state_copy = self.cast(state, copy=True)
        new_state = (1 - channel.coefficient_sum) * state
        for coeff, gate in zip(channel.coefficients, channel.gates):
            state = self.apply_gate_density_matrix(gate, state, nqubits)
            new_state += coeff * state
            # reset the state
            if not channel._all_unitary_operators:
                state = self.cast(state_copy, copy=True)
            else:
                state = self.apply_gate_density_matrix(
                    gate, state, nqubits, inverse=True
                )
        return new_state

    def collapse_state(self, state, qubits, shot, nqubits, normalize=True):
        state = self.cast(state)
        qubits = self.cast([nqubits - q - 1 for q in reversed(qubits)], dtype="int32")
        if normalize:
            return self.ops.collapse_state_normalized(state, qubits, int(shot), nqubits)
        else:
            return self.ops.collapse_state(state, qubits, int(shot), nqubits)

    def collapse_density_matrix(self, state, qubits, shot, nqubits, normalize=True):
        state = self.cast(state)
        shape = state.shape
        dm_qubits = [q + nqubits for q in qubits]
        state = self.collapse_state(state.ravel(), dm_qubits, shot, 2 * nqubits, False)
        state = self.collapse_state(state, qubits, shot, 2 * nqubits, False)
        state = self.np.reshape(state, shape)
        if normalize:
            state = state / self.np.trace(state)
        return state

    # def calculate_probabilities(self, state, qubits, nqubits): Inherited from ``NumpyBackend``

    # def sample_shots(self, probabilities, nshots): Inherited from ``NumpyBackend``

    # def aggregate_shots(self, shots): Inherited from ``NumpyBackend``

    # def samples_to_binary(self, samples, nqubits): Inherited from ``NumpyBackend``

    # def samples_to_decimal(self, samples, nqubits): Inherited from ``NumpyBackend``

    def sample_frequencies(self, probabilities, nshots):
        from qibo.config import SHOT_METROPOLIS_THRESHOLD

        if nshots < SHOT_METROPOLIS_THRESHOLD:
            return super().sample_frequencies(probabilities, nshots)

        import collections

        seed = np.random.randint(0, int(1e8), dtype="int64")
        nqubits = int(np.log2(tuple(probabilities.shape)[0]))
        frequencies = np.zeros(2**nqubits, dtype="int64")
        # always fall back to numba CPU backend because for ops not implemented on GPU
        frequencies = self.measure_frequencies_op(
            frequencies, probabilities, nshots, nqubits, seed, self.nthreads
        )
        return collections.Counter({i: f for i, f in enumerate(frequencies) if f > 0})

    # def calculate_frequencies(self, samples): Inherited from ``NumpyBackend``

    # def assert_allclose(self, value, target, rtol=1e-7, atol=0.0): Inherited from ``NumpyBackend``



# This file contains the C++/CUDA implementation of many methods
# defined in gates.py and ops.py.
# The methods in gates.py and ops.py are called by NumbaBackend
# while the functions and kernels here are called by CupyBackend.

# ---------- HELPER METHODS ----------

multicontrol_index = """
// C++ implementation of gates.py:multicontrol_index()
__device__ long multicontrol_index(const int* qubits, long g, int ncontrols) {
  long i = g;
  for (int iq = 0; iq < ncontrols; iq++) {
      const int n = qubits[iq];
      long k = ((long)1 << n);
      i = ((long)((long)i >> n) << (n + 1)) + (i & (k - 1)) + k;
  }
  return i;
}
"""  # pragma: no cover


_apply_gate = """
// Helper method for apply_gate_kernel()
__device__ void _apply_gate(T& state1, T& state2, const T* gate) {
  const T buffer = state1;
  state1 = gate[0] * state1 + gate[1] * state2;
  state2 = gate[2] * buffer + gate[3] * state2;
}
"""  # pragma: no cover


_apply_x = """
// Helper method for apply_x_kernel()
__device__ void _apply_x(T& state1, T& state2) {
  const T buffer = state1;
  state1 = state2;
  state2 = buffer;
}
"""  # pragma: no cover


_apply_y = """
// Helper method for apply_y_kernel()
__device__ void _apply_y(T& state1, T& state2) {
  state1 = state1 * T(0, 1);
  state2 = state2 * T(0, -1);
  const T buffer = state1;
  state1 = state2;
  state2 = buffer;
}
"""  # pragma: no cover


_apply_z = """
// Helper method for apply_z_kernel()
__device__ void _apply_z(T& state) {
  state = state * T(-1);
}
"""  # pragma: no cover


_apply_z_pow = """
// Helper method for apply_z_pow_kernel()
__device__ void _apply_z_pow(T& state, T gate) {
  state = state * gate;
}
"""  # pragma: no cover


_apply_two_qubit_gate = """
// Helper method for apply_two_qubit_gate_kernel()
__device__ void _apply_two_qubit_gate(T& state0, T& state1, T& state2, T& state3,
                                      const T* gate) {
  const T buffer0 = state0;
  const T buffer1 = state1;
  const T buffer2 = state2;
  state0 = gate[0]  * state0  + gate[1]  * state1
         + gate[2]  * state2  + gate[3]  * state3;
  state1 = gate[4]  * buffer0 + gate[5]  * state1
         + gate[6]  * state2  + gate[7]  * state3;
  state2 = gate[8]  * buffer0 + gate[9]  * buffer1
         + gate[10] * state2  + gate[11] * state3;
  state3 = gate[12] * buffer0 + gate[13] * buffer1
         + gate[14] * buffer2 + gate[15] * state3;
}
"""  # pragma: no cover


_apply_fsim = """
// Helper method for apply_fsim_kernel()
__device__ void _apply_fsim(T& state1, T& state2, T& state3, const T* gate) {
  const T buffer = state1;
  state1 = gate[0] * state1 + gate[1] * state2;
  state2 = gate[2] * buffer + gate[3] * state2;
  state3 = gate[4] * state3;
}
"""  # pragma: no cover


multitarget_index = """
// C++ implementation of gates.py:multitarget_index()
__device__ long multitarget_index(const long* targets, long i, int ntargets) {
  long t = 0;
  for (int u = 0; u < ntargets; u++) {
    t += ((long)(i >> u) & 1) * targets[u];
  }
  return t;
}
"""  # pragma: no cover


collapse_index = """
// C++ implementation of ops.py:collapse_index()
__device__ long collapse_index(const int* qubits, long g, long h, int ntargets) {
  long i = g;
  for (auto iq = 0; iq < ntargets; iq++) {
    const auto n = qubits[iq];
    long k = (long)1 << n;
    i = ((long)((long)i >> n) << (n + 1)) + (i & (k - 1));
    i += ((long)((long)(h >> iq) % 2) * k);
  }
  return i;
}
"""  # pragma: no cover

# ---------- KERNELS ----------

apply_gate_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_gate}"""
    + """
// C++ implementation of gates.py:apply_gate_kernel()
extern "C"
__global__ void apply_gate_kernel(T* state, long tk, int m, const T* gate) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = ((long)((long)g >> m) << (m + 1)) + (g & (tk - 1));
  _apply_gate(state[i], state[i + tk], gate);
}
"""
)  # pragma: no cover


apply_x_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_x}"""
    + """
// C++ implementation of gates.py:apply_x_kernel()
extern "C"
__global__ void apply_x_kernel(T* state, long tk, int m) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = ((long)((long)g >> m) << (m + 1)) + (g & (tk - 1));
  _apply_x(state[i], state[i + tk]);
}
"""
)  # pragma: no cover


apply_y_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_y}"""
    + """
// C++ implementation of gates.py:apply_y_kernel()
extern "C"
__global__ void apply_y_kernel(T* state, long tk, int m) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = ((long)((long)g >> m) << (m + 1)) + (g & (tk - 1));
  _apply_y(state[i], state[i + tk]);
}
"""
)  # pragma: no cover


apply_z_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_z}"""
    + """
// C++ implementation of gates.py:apply_z_kernel()
extern "C"
__global__ void apply_z_kernel(T* state, long tk, int m) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = ((long)((long)g >> m) << (m + 1)) + (g & (tk - 1));
  _apply_z(state[i + tk]);
}
"""
)  # pragma: no cover


apply_z_pow_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_z_pow}"""
    + """
// C++ implementation of gates.py:apply_z_pow_kernel()
extern "C"
__global__ void apply_z_pow_kernel(T* state, long tk, int m,
                                   const T* gate) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = ((long)((long)g >> m) << (m + 1)) + (g & (tk - 1));
  _apply_z_pow(state[i + tk], gate[0]);
}
"""
)  # pragma: no cover


apply_two_qubit_gate_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_two_qubit_gate}"""
    + """
// C++ implementation of gates.py:apply_two_qubit_gate_kernel()
// the portion of code before the parallel for of the Python
// method is in backends.py:CupyBackend.two_qubit_base()
extern "C"
__global__ void apply_two_qubit_gate_kernel(T* state, long tk1, long tk2,
                                            int m1, int m2, long uk1, long uk2,
                                            const T* gate) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  long i = ((long)((long)g >> m1) << (m1 + 1)) + (g & (tk1 - 1));
  i = ((long)((long)i >> m2) << (m2 + 1)) + (i & (tk2 - 1));
  _apply_two_qubit_gate(state[i], state[i + uk1], state[i + uk2], state[i + uk1 + uk2], gate);
}
"""
)  # pragma: no cover


apply_fsim_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_fsim}"""
    + """
// C++ implementation of gates.py:apply_fsim_kernel()
// the portion of code before the parallel for of the Python
// method is in backends.py:CupyBackend.two_qubit_base()
extern "C"
__global__ void apply_fsim_kernel(T* state, long tk1, long tk2,
                                  int m1, int m2, long uk1, long uk2,
                                  const T* gate) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  long i = ((long)((long)g >> m1) << (m1 + 1)) + (g & (tk1 - 1));
  i = ((long)((long)i >> m2) << (m2 + 1)) + (i & (tk2 - 1));
  _apply_fsim(state[i + uk1], state[i + uk2], state[i + uk1 + uk2], gate);
}
"""
)  # pragma: no cover


apply_swap_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_x}"""
    + """
// C++ implementation of gates.py:apply_swap_kernel()
extern "C"
__global__ void apply_swap_kernel(T* state, long tk1, long tk2,
                                  int m1, int m2, long uk1, long uk2) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  long i = ((long)((long)g >> m1) << (m1 + 1)) + (g & (tk1 - 1));
  i = ((long)((long)i >> m2) << (m2 + 1)) + (i & (tk2 - 1));
  _apply_x(state[i + tk2], state[i + tk1]);
}
"""
)  # pragma: no cover


multicontrol_apply_gate_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_gate}
{multicontrol_index}"""
    + """
// C++ implementation of gates.py:multicontrol_apply_gate_kernel()
extern "C"
__global__ void multicontrol_apply_gate_kernel(T* state, long tk, int m, const T* gate,
                                               const int* qubits, int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = multicontrol_index(qubits, g, ncontrols);
  _apply_gate(state[i - tk], state[i], gate);
}
"""
)  # pragma: no cover


multicontrol_apply_x_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_x}
{multicontrol_index}"""
    + """
// C++ implementation of gates.py:multicontrol_apply_x_kernel()
extern "C"
__global__ void multicontrol_apply_x_kernel(T* state, long tk, int m,
                                            const int* qubits, int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = multicontrol_index(qubits, g, ncontrols);
  _apply_x(state[i - tk], state[i]);
}
"""
)  # pragma: no cover


multicontrol_apply_y_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_y}
{multicontrol_index}"""
    + """
// C++ implementation of gates.py:multicontrol_apply_y_kernel()
extern "C"
__global__ void multicontrol_apply_y_kernel(T* state, long tk, int m,
                                            const int* qubits, int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = multicontrol_index(qubits, g, ncontrols);
  _apply_y(state[i - tk], state[i]);
}
"""
)  # pragma: no cover


multicontrol_apply_z_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_z}
{multicontrol_index}"""
    + """
// C++ implementation of gates.py:multicontrol_apply_z_kernel()
extern "C"
__global__ void multicontrol_apply_z_kernel(T* state, long tk, int m,
                                            const int* qubits, int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = multicontrol_index(qubits, g, ncontrols);
  _apply_z(state[i]);
}
"""
)  # pragma: no cover


multicontrol_apply_z_pow_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_z_pow}
{multicontrol_index}"""
    + """
// C++ implementation of gates.py:multicontrol_apply_z_pow_kernel()
extern "C"
__global__ void multicontrol_apply_z_pow_kernel(T* state, long tk, int m,
                                                const T* gate,
                                                const int* qubits, int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = multicontrol_index(qubits, g, ncontrols);
  _apply_z_pow(state[i], gate[0]);
}
"""
)  # pragma: no cover


multicontrol_apply_two_qubit_gate_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_two_qubit_gate}
{multicontrol_index}"""
    + """
// C++ implementation of gates.py:multicontrol_apply_two_qubit_gate_kernel()
// the portion of code before the parallel for of the Python method
// is in backends.py:CupyBackend.two_qubit_base()
extern "C"
__global__ void multicontrol_apply_two_qubit_gate_kernel(T* state,
                                                         long tk1, long tk2,
                                                         int m1, int m2,
                                                         long uk1, long uk2,
                                                         const T* gate,
                                                         const int* qubits,
                                                         int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = multicontrol_index(qubits, g, ncontrols);
  _apply_two_qubit_gate(state[i - uk1 - uk2], state[i - uk2], state[i - uk1], state[i], gate);
}
"""
)  # pragma: no cover


multicontrol_apply_fsim_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_fsim}
{multicontrol_index}"""
    + """
// C++ implementation of gates.py:multicontrol_apply_fsim_kernel()
// the portion of code before the parallel for of the Python method
// is in backends.py:CupyBackend.two_qubit_base()
extern "C"
__global__ void multicontrol_apply_fsim_kernel(T* state,
                                               long tk1, long tk2,
                                               int m1, int m2,
                                               long uk1, long uk2,
                                               const T* gate,
                                               const int* qubits,
                                               int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = multicontrol_index(qubits, g, ncontrols);
  _apply_fsim(state[i - uk2], state[i - uk1], state[i], gate);
}
"""
)  # pragma: no cover


multicontrol_apply_swap_kernel = (
    f"""
#include <cupy/complex.cuh>
{_apply_x}
{multicontrol_index}"""
    + """
// C++ implementation of gates.py:multicontrol_apply_swap_kernel()
extern "C"
__global__ void multicontrol_apply_swap_kernel(T* state,
                                               long tk1, long tk2,
                                               int m1, int m2,
                                               long uk1, long uk2,
                                               const int* qubits,
                                               int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long i = multicontrol_index(qubits, g, ncontrols);
  _apply_x(state[i - tk1], state[i - tk2]);
}
"""
)  # pragma: no cover


apply_multi_qubit_gate_kernel = (
    f"""
#include <cupy/complex.cuh>
{multicontrol_index}
{multitarget_index}"""
    + """
// C++ implementation of gates.py:apply_multi_qubit_gate_kernel()
extern "C" __global__ void
__launch_bounds__(MAX_BLOCK_SIZE) // to prevent cuda_error_launch_out_of_resources.
                                  // the maximum block size is chosen in backends.py
                                  // and it is replaced here before compilation.
apply_multi_qubit_gate_kernel(T* state,
                              const T* gate,
                              const int* qubits,
                              const long* targets,
                              int ntargets,
                              int ncontrols) {
  const long g = blockIdx.x * blockDim.x + threadIdx.x;
  const long ig = multicontrol_index(qubits, g, ncontrols);
  T buffer[nsubstates];
  for (auto i = 0; i < nsubstates; i++) {
    const long t = ig - multitarget_index(targets, nsubstates - i - 1, ntargets);
    buffer[i] = state[t];
  }
  for (auto i = 0; i < nsubstates; i++) {
    const long t = ig - multitarget_index(targets, nsubstates - i - 1, ntargets);
    T new_state_elem = T(0., 0.); // use local variable because it is faster than global ones
    for (auto j = 0; j < nsubstates; j++) {
      new_state_elem += gate[nsubstates * i + j] * buffer[j];
    }
    state[t] = new_state_elem;
  }
}
"""
)  # pragma: no cover


collapse_state_kernel = (
    f"""
#include <cupy/complex.cuh>
{collapse_index}"""
    + """
// C++ implementation of ops.py:collapse_state() and ops.py:collapse_state_normalized()
// Only the parallel for is implemented here. the other portions of code are
// implemented in backends.py:CupyBackend.collapse_state()
extern "C"
__global__ void collapse_state_kernel(T* state, const int* qubits,
                                      const long result, int ntargets) {
  const auto g = blockIdx.x * blockDim.x + threadIdx.x;
  const long nsubstates = (long)1 << ntargets;
  for (auto h = 0; h < result; h++) {
    state[collapse_index(qubits, g, h, ntargets)] = T(0, 0);
  }
  for (auto h = result + 1; h < nsubstates; h++) {
    state[collapse_index(qubits, g, h, ntargets)] = T(0, 0);
  }
}
"""
)  # pragma: no cover


initial_state_kernel = """
#include <cupy/complex.cuh>

// C++ implementation of ops.py:initial_state_vector()
// In contrast to the Python method, the state is inizialized
// to zero in backends.py:CupyBackend.initial_state, then a
// single thread execute this kernel and set the first element to 1
extern "C" __global__ void initial_state_kernel(T* state) {
  state[0] = T(1, 0);
}
"""  # pragma: no cover



import numpy as np
from qibo.backends.numpy import NumpyBackend
from qibo.config import log, raise_error

from qibojit.backends.cpu import NumbaBackend
from qibojit.backends.matrices import CupyMatrices, CuQuantumMatrices, CustomMatrices


class CupyBackend(NumbaBackend):  # pragma: no cover
    # CI does not have GPUs

    DEFAULT_BLOCK_SIZE = 1024
    MAX_NUM_TARGETS = 7

    def __init__(self):
        NumpyBackend.__init__(self)

        import cupy as cp  # pylint: disable=import-error
        import cupy_backends  # pylint: disable=import-error

        self.name = "qibojit"
        self.platform = "cupy"
        self.versions["cupy"] = cp.__version__

        self.supports_multigpu = True
        self.numeric_types = (
            int,
            float,
            complex,
            cp.int32,
            cp.int64,
            cp.float32,
            cp.float64,
            cp.complex64,
            cp.complex128,
        )
        self.tensor_types = (np.ndarray, cp.ndarray)
        from scipy import sparse

        self.npsparse = sparse
        self.sparse = cp.sparse
        self.device = "/GPU:0"
        self.kernel_type = "double"
        self.matrices = CupyMatrices(self.dtype)
        self.custom_matrices = CustomMatrices(self.dtype)
        try:
            if not cp.cuda.runtime.getDeviceCount():  # pragma: no cover
                raise RuntimeError("Cannot use cupy backend if GPU is not available.")
        except cp.cuda.runtime.CUDARuntimeError:
            raise ImportError("Could not detect cupy compatible devices.")

        self.cp = cp
        self.is_hip = cupy_backends.cuda.api.runtime.is_hip
        self.KERNELS = (
            "apply_gate",
            "apply_x",
            "apply_y",
            "apply_z",
            "apply_z_pow",
            "apply_two_qubit_gate",
            "apply_fsim",
            "apply_swap",
        )

        # load core kernels
        self.gates = {}
        from qibojit.custom_operators import raw_kernels

        def kernel_loader(name, ktype):
            code = getattr(raw_kernels, name)
            code = code.replace("T", f"thrust::complex<{ktype}>")
            gate = cp.RawKernel(code, name, ("--std=c++11",))
            self.gates[f"{name}_{ktype}"] = gate

        for ktype in ("float", "double"):
            for name in self.KERNELS:
                kernel_loader(f"{name}_kernel", ktype)
                kernel_loader(f"multicontrol_{name}_kernel", ktype)
            kernel_loader("collapse_state_kernel", ktype)
            kernel_loader("initial_state_kernel", ktype)

        # load multiqubit kernels
        name = "apply_multi_qubit_gate_kernel"
        for ntargets in range(3, self.MAX_NUM_TARGETS + 1):
            for ktype in ("float", "double"):
                code = getattr(raw_kernels, name)
                code = code.replace("T", f"thrust::complex<{ktype}>")
                code = code.replace("nsubstates", str(2**ntargets))
                code = code.replace("MAX_BLOCK_SIZE", str(self.DEFAULT_BLOCK_SIZE))
                gate = cp.RawKernel(code, name, ("--std=c++11",))
                self.gates[f"{name}_{ktype}_{ntargets}"] = gate

        # load numba op for measuring frequencies
        from qibojit.custom_operators.ops import measure_frequencies

        self.measure_frequencies_op = measure_frequencies

        # number of available GPUs (for multigpu)
        self.ngpus = cp.cuda.runtime.getDeviceCount()

    def set_precision(self, precision):
        super().set_precision(precision)
        if self.dtype == "complex128":
            self.kernel_type = "double"
        elif self.dtype == "complex64":
            self.kernel_type = "float"

    def set_device(self, device):
        if "GPU" not in device:
            raise_error(
                ValueError, f"Device {device} is not available for {self} backend."
            )
        # TODO: Raise error if GPU is not available
        self.device = device

    def cast(self, x, dtype=None, copy=False):
        if dtype is None:
            dtype = self.dtype
        if self.sparse.issparse(x):
            if dtype != x.dtype:
                return x.astype(dtype)
            else:
                return x
        elif self.npsparse.issparse(x):
            cls = getattr(self.sparse, x.__class__.__name__)
            return cls(x, dtype=dtype)
        elif isinstance(x, self.cp.ndarray) and copy:
            return self.cp.copy(self.cp.asarray(x, dtype=dtype))
        else:
            return self.cp.asarray(x, dtype=dtype)

    def to_numpy(self, x):
        if isinstance(x, self.cp.ndarray):
            return x.get()
        elif isinstance(x, list):
            return self.cp.asarray(x).get()
        elif self.sparse.issparse(x):
            return x.toarray().get()
        elif self.npsparse.issparse(x):
            return x.toarray()
        return np.array(x, copy=False)

    def is_sparse(self, x):
        return self.sparse.issparse(x) or self.npsparse.issparse(x)

    def zero_state(self, nqubits):
        n = 1 << nqubits
        kernel = self.gates.get(f"initial_state_kernel_{self.kernel_type}")
        state = self.cp.zeros(n, dtype=self.dtype)
        kernel((1,), (1,), [state])
        self.cp.cuda.stream.get_current_stream().synchronize()
        return state

    def zero_density_matrix(self, nqubits):
        n = 1 << nqubits
        kernel = self.gates.get(f"initial_state_kernel_{self.kernel_type}")
        state = self.cp.zeros(n * n, dtype=self.dtype)
        kernel((1,), (1,), [state])
        self.cp.cuda.stream.get_current_stream().synchronize()
        return state.reshape((n, n))

    def identity_density_matrix(self, nqubits, normalize: bool = True):
        n = 1 << nqubits
        state = self.cp.eye(n, dtype=self.dtype)
        self.cp.cuda.stream.get_current_stream().synchronize()
        if normalize:
            state /= 2**nqubits
        return state.reshape((n, n))

    def plus_state(self, nqubits):
        state = self.cp.ones(2**nqubits, dtype=self.dtype)
        state /= self.cp.sqrt(2**nqubits)
        return state

    def plus_density_matrix(self, nqubits):
        state = self.cp.ones(2 * (2**nqubits,), dtype=self.dtype)
        state /= 2**nqubits
        return state

    def matrix_fused(self, gate):
        npmatrix = super().matrix_fused(gate)
        return self.cast(npmatrix, dtype=self.dtype)

    # def control_matrix(self, gate): Inherited from ``NumpyBackend``

    def calculate_blocks(self, nstates, block_size=DEFAULT_BLOCK_SIZE):
        """Compute the number of blocks and of threads per block.

        The total number of threads is always equal to ``nstates``, give that
        the kernels are designed to execute only one out of ``nstates`` updates.
        Therefore, the number of threads per block (``block_size``) changes also
        the total number of blocks. By default, it is set to ``self.DEFAULT_BLOCK_SIZE``.
        """
        # Compute the number of blocks so that at least ``nstates`` threads are launched
        nblocks = (nstates + block_size - 1) // block_size
        if nstates < block_size:
            nblocks = 1
            block_size = nstates
        return nblocks, block_size

    def one_qubit_base(self, state, nqubits, target, kernel, gate, qubits):
        ncontrols = len(qubits) - 1 if qubits is not None else 0
        m = nqubits - target - 1
        tk = 1 << m
        nstates = 1 << (nqubits - ncontrols - 1)
        if kernel in ("apply_x", "apply_y", "apply_z"):
            args = (state, tk, m)
        else:
            args = (state, tk, m, gate)

        if ncontrols:
            kernel = self.gates.get(f"multicontrol_{kernel}_kernel_{self.kernel_type}")
            args += (qubits, ncontrols + 1)
        else:
            kernel = self.gates.get(f"{kernel}_kernel_{self.kernel_type}")

        nblocks, block_size = self.calculate_blocks(nstates)
        kernel((nblocks,), (block_size,), args)
        self.cp.cuda.stream.get_current_stream().synchronize()
        return state

    def two_qubit_base(self, state, nqubits, target1, target2, kernel, gate, qubits):
        ncontrols = len(qubits) - 2 if qubits is not None else 0
        if target1 > target2:
            m1 = nqubits - target1 - 1
            m2 = nqubits - target2 - 1
            tk1, tk2 = 1 << m1, 1 << m2
            uk1, uk2 = tk2, tk1
        else:
            m1 = nqubits - target2 - 1
            m2 = nqubits - target1 - 1
            tk1, tk2 = 1 << m1, 1 << m2
            uk1, uk2 = tk1, tk2
        nstates = 1 << (nqubits - 2 - ncontrols)

        if kernel == "apply_swap":
            args = (state, tk1, tk2, m1, m2, uk1, uk2)
        else:
            args = (state, tk1, tk2, m1, m2, uk1, uk2, gate)
            assert state.dtype == args[-1].dtype

        if ncontrols:
            kernel = self.gates.get(f"multicontrol_{kernel}_kernel_{self.kernel_type}")
            args += (qubits, ncontrols + 2)
        else:
            kernel = self.gates.get(f"{kernel}_kernel_{self.kernel_type}")

        nblocks, block_size = self.calculate_blocks(nstates)
        kernel((nblocks,), (block_size,), args)
        self.cp.cuda.stream.get_current_stream().synchronize()
        return state

    def multi_qubit_base(self, state, nqubits, targets, gate, qubits):
        assert gate is not None
        if qubits is None:
            qubits = self.cast(
                sorted(nqubits - q - 1 for q in targets), dtype=self.cp.int32
            )
        ntargets = len(targets)
        if ntargets > self.MAX_NUM_TARGETS:
            raise ValueError(
                f"Number of target qubits must be <= {self.MAX_NUM_TARGETS}"
                f" but is {ntargets}."
            )
        nactive = len(qubits)
        targets = self.cp.asarray(
            tuple(1 << (nqubits - t - 1) for t in targets[::-1]), dtype=self.cp.int64
        )
        nstates = 1 << (nqubits - nactive)
        nsubstates = 1 << ntargets
        nblocks, block_size = self.calculate_blocks(nstates)
        kernel = self.gates.get(
            f"apply_multi_qubit_gate_kernel_{self.kernel_type}_{ntargets}"
        )
        args = (state, gate, qubits, targets, ntargets, nactive)
        kernel((nblocks,), (block_size,), args)
        self.cp.cuda.stream.get_current_stream().synchronize()
        return state

    def _create_qubits_tensor(self, gate, nqubits):
        qubits = super()._create_qubits_tensor(gate, nqubits)
        return self.cp.asarray(qubits, dtype=self.cp.int32)

    def _as_custom_matrix(self, gate):
        from qibo.gates import Unitary

        if isinstance(gate, Unitary):
            matrix = gate.parameters[0]
            if isinstance(matrix, self.cp.ndarray):
                return matrix.ravel()

        matrix = super()._as_custom_matrix(gate)
        return self.cp.asarray(matrix.ravel())

    # def apply_gate(self, gate, state, nqubits): Inherited from ``NumbaBackend``

    # def apply_gate_density_matrix(self, gate, state, nqubits, inverse=False): Inherited from ``NumbaBackend``

    # def _apply_ygate_density_matrix(self, gate, state, nqubits): Inherited from ``NumbaBackend``

    # def apply_channel(self, gate): Inherited from ``NumbaBackend``

    # def apply_channel_density_matrix(self, channel, state, nqubits): Inherited from ``NumbaBackend``

    def collapse_state(self, state, qubits, shot, nqubits, normalize=True):
        ntargets = len(qubits)
        nstates = 1 << (nqubits - ntargets)
        nblocks, block_size = self.calculate_blocks(nstates)

        state = self.cast(state)
        qubits = self.cast(
            [nqubits - q - 1 for q in reversed(qubits)], dtype=self.cp.int32
        )
        args = [state, qubits, int(shot), ntargets]
        kernel = self.gates.get(f"collapse_state_kernel_{self.kernel_type}")
        kernel((nblocks,), (block_size,), args)
        self.cp.cuda.stream.get_current_stream().synchronize()

        if normalize:
            norm = self.cp.sqrt(self.cp.sum(self.cp.square(self.cp.abs(state))))
            state = state / norm
        return state

    # def collapse_density_matrix(self, state, qubits, shot, nqubits, normalize=True): Inherited from ``NumbaBackend``

    # def reset_error_density_matrix(self, gate, state, nqubits): Inherited from ``NumpyBackend``

    def execute_distributed_circuit(
        self,
        circuit,
        initial_state=None,
        nshots=1000,
    ):
        import joblib
        from qibo.gates import M
        from qibo.result import CircuitResult, MeasurementOutcomes, QuantumState

        if not circuit.queues.queues:
            circuit.queues.set(circuit.queue)

        try:
            cpu_backend = NumbaBackend()
            cpu_backend.set_precision(self.precision)
            ops = MultiGpuOps(self, cpu_backend, circuit)

            if initial_state is None:
                # Generate pieces for |000...0> state
                pieces = [cpu_backend.zero_state(circuit.nlocal)]
                pieces.extend(
                    np.zeros(2**circuit.nlocal, dtype=self.dtype)
                    for _ in range(circuit.ndevices - 1)
                )
            elif isinstance(initial_state, CircuitResult) or isinstance(
                initial_state, QuantumState
            ):
                # TODO: Implement this
                if isinstance(initial_state.execution_result, list):
                    pieces = initial_state.execution_result
                else:
                    pieces = ops.to_pieces(initial_state.state())
            elif isinstance(initial_state, self.tensor_types):
                pieces = ops.to_pieces(initial_state)
            else:
                raise_error(
                    TypeError,
                    "Initial state type {} is not supported by "
                    "distributed circuits.".format(type(initial_state)),
                )
            for gate in circuit.queue:
                if isinstance(gate, M):
                    gate.result.backend = CupyBackend()
            special_gates = iter(circuit.queues.special_queue)
            for i, queues in enumerate(circuit.queues.queues):
                if queues:  # standard gate
                    config = circuit.queues.device_to_ids.items()
                    pool = joblib.Parallel(n_jobs=circuit.ndevices, prefer="threads")
                    pool(
                        joblib.delayed(ops.apply_gates)(pieces, queues, ids, device)
                        for device, ids in config
                    )

                else:  # special gate
                    gate = next(special_gates)
                    if isinstance(gate, tuple):  # SWAP global-local qubit
                        global_qubit, local_qubit = gate
                        pieces = ops.swap(pieces, global_qubit, local_qubit)
                    else:
                        pieces = ops.apply_special_gate(pieces, gate)

            for gate in special_gates:  # pragma: no cover
                pieces = ops.apply_special_gate(pieces, gate)

            state = ops.to_tensor(pieces)

            if circuit.has_unitary_channel:
                # here we necessarily have `density_matrix=True`, otherwise
                # execute_circuit_repeated would have been called
                if circuit.measurements:
                    circuit._final_state = CircuitResult(
                        state, circuit.measurements, self, nshots=nshots
                    )
                    return circuit._final_state
                else:
                    circuit._final_state = QuantumState(state, self)
                    return circuit._final_state
            else:
                if circuit.measurements:
                    circuit._final_state = CircuitResult(
                        state, circuit.measurements, self, nshots=nshots
                    )
                    return circuit._final_state
                else:
                    circuit._final_state = QuantumState(state, self)
                    return circuit._final_state

        except self.oom_error:
            raise_error(
                RuntimeError,
                "State does not fit in memory during distributed "
                "execution. Please create a new circuit with "
                "different device configuration and try again.",
            )

    # def calculate_symbolic(self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20): Inherited from ``NumpyBackend``

    # def calculate_symbolic_density_matrix(self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20): Inherited from ``NumpyBackend``

    def calculate_probabilities(self, state, qubits, nqubits):
        try:
            probs = super().calculate_probabilities(state, qubits, nqubits)
        except MemoryError:
            # fall back to CPU
            probs = super().calculate_probabilities(
                self.to_numpy(state), qubits, nqubits
            )
        return probs

    def sample_shots(self, probabilities, nshots):
        # Sample shots on CPU
        probabilities = self.to_numpy(probabilities)
        return super().sample_shots(probabilities, nshots)

    # def aggregate_shots(self, shots): Inherited from ``NumpyBackend``

    # def samples_to_binary(self, samples, nqubits): Inherited from ``NumpyBackend``

    # def samples_to_decimal(self, samples, nqubits): Inherited from ``NumpyBackend``

    def sample_frequencies(self, probabilities, nshots):
        # Sample frequencies on CPU
        probabilities = self.to_numpy(probabilities)
        return super().sample_frequencies(probabilities, nshots)

    # def calculate_frequencies(self, samples): Inherited from ``NumpyBackend``

    # def assert_allclose(self, value, target, rtol=1e-7, atol=0.0): Inherited from ``NumpyBackend``

    def calculate_expectation_state(self, matrix, state, normalize):
        state = self.cast(state)
        statec = self.cp.conj(state)
        hstate = matrix @ state
        ev = self.cp.real(self.cp.sum(statec * hstate))
        if normalize:
            norm = self.cp.sum(self.cp.square(self.cp.abs(state)))
            ev = ev / norm
        return ev

    def calculate_expectation_density_matrix(self, matrix, state, normalize):
        state = self.cast(state)
        ev = self.cp.real(self.cp.trace(matrix @ state))
        if normalize:
            norm = self.cp.real(self.cp.trace(state))
            ev = ev / norm
        return ev

    def calculate_eigenvalues(self, matrix, k=6):
        if self.is_sparse(matrix):
            log.warning(
                "Calculating sparse matrix eigenvectors because "
                "sparse modules do not provide ``eigvals`` method."
            )
            return self.calculate_eigenvectors(matrix, k=k)[0]
        return self.cp.linalg.eigvalsh(matrix)

    def calculate_eigenvectors(self, matrix, k=6):
        if self.is_sparse(matrix):
            if k < matrix.shape[0]:
                # Fallback to numpy because cupy's ``sparse.eigh`` does not support 'SA'
                from scipy.sparse.linalg import eigsh  # pylint: disable=import-error

                result = eigsh(matrix.get(), k=k, which="SA")
                return self.cast(result[0]), self.cast(result[1])
            matrix = matrix.toarray()
        if self.is_hip:
            # Fallback to numpy because eigh is not implemented in rocblas
            result = self.np.linalg.eigh(self.to_numpy(matrix))
            return self.cast(result[0]), self.cast(result[1])
        else:
            return self.cp.linalg.eigh(matrix)

    def calculate_matrix_exp(self, a, matrix, eigenvectors=None, eigenvalues=None):
        if eigenvectors is None or self.is_sparse(matrix):
            if self.is_sparse(matrix):
                from scipy.sparse.linalg import expm

                return self.cast(expm(-1j * a * matrix.get()))
            else:
                from cupyx.scipy.linalg import expm  # pylint: disable=import-error

                return self.cast(expm(-1j * a * matrix))
        else:
            expd = self.cp.diag(self.cp.exp(-1j * a * eigenvalues))
            ud = self.cp.transpose(self.cp.conj(eigenvectors))
            return self.cp.matmul(eigenvectors, self.cp.matmul(expd, ud))


class CuQuantumBackend(CupyBackend):  # pragma: no cover
    # CI does not test for GPU

    def __init__(self):
        super().__init__()
        import cuquantum  # pylint: disable=import-error
        from cuquantum import custatevec as cusv  # pylint: disable=import-error

        self.cuquantum = cuquantum
        self.cusv = cusv
        self.platform = "cuquantum"
        self.versions["cuquantum"] = self.cuquantum.__version__
        self.supports_multigpu = True
        self.handle = self.cusv.create()
        self.custom_matrices = CuQuantumMatrices(self.dtype)

    def __del__(self):
        if hasattr(self, "cusv"):
            self.cusv.destroy(self.handle)

    def set_precision(self, precision):
        if precision != self.precision:
            super().set_precision(precision)
            if self.custom_matrices:
                self.custom_matrices = CuQuantumMatrices(self.dtype)

    def get_cuda_type(self, dtype="complex64"):
        if dtype == "complex128":
            return (
                self.cuquantum.cudaDataType.CUDA_C_64F,
                self.cuquantum.ComputeType.COMPUTE_64F,
            )
        elif dtype == "complex64":
            return (
                self.cuquantum.cudaDataType.CUDA_C_32F,
                self.cuquantum.ComputeType.COMPUTE_32F,
            )
        else:
            raise TypeError("Type can be either complex64 or complex128")

    def one_qubit_base(self, state, nqubits, target, kernel, gate, qubits=None):
        ntarget = 1
        target = nqubits - target - 1
        if qubits is not None:
            qubits = self.to_numpy(qubits)
            ncontrols = len(qubits) - 1
            controls = self.np.asarray(
                [i for i in qubits if i != target], dtype="int32"
            )
        else:
            ncontrols = 0
            controls = self.np.empty(0)
        adjoint = 0
        target = self.np.asarray([target], dtype=self.np.int32)

        state = self.cast(state)
        gate = self.cast(gate)
        assert state.dtype == gate.dtype
        data_type, compute_type = self.get_cuda_type(state.dtype)
        if isinstance(gate, self.cp.ndarray):
            gate_ptr = gate.data.ptr
        elif isinstance(gate, self.np.ndarray):
            gate_ptr = gate.ctypes.data
        else:
            raise ValueError

        workspaceSize = self.cusv.apply_matrix_get_workspace_size(
            self.handle,
            data_type,
            nqubits,
            gate_ptr,
            data_type,
            self.cusv.MatrixLayout.ROW,
            adjoint,
            ntarget,
            ncontrols,
            compute_type,
        )

        # check the size of external workspace
        if workspaceSize > 0:
            workspace = self.cp.cuda.memory.alloc(workspaceSize)
            workspace_ptr = workspace.ptr
        else:
            workspace_ptr = 0

        self.cusv.apply_matrix(
            self.handle,
            state.data.ptr,
            data_type,
            nqubits,
            gate_ptr,
            data_type,
            self.cusv.MatrixLayout.ROW,
            adjoint,
            target.ctypes.data,
            ntarget,
            controls.ctypes.data,
            0,
            ncontrols,
            compute_type,
            workspace_ptr,
            workspaceSize,
        )

        return state

    def two_qubit_base(
        self, state, nqubits, target1, target2, kernel, gate, qubits=None
    ):
        ntarget = 2
        target1 = nqubits - target1 - 1
        target2 = nqubits - target2 - 1
        target = self.np.asarray([target2, target1], dtype=self.np.int32)
        if qubits is not None:
            ncontrols = len(qubits) - 2
            qubits = self.to_numpy(qubits)
            controls = self.np.asarray(
                [i for i in qubits if i not in [target1, target2]], dtype=self.np.int32
            )
        else:
            ncontrols = 0
            controls = self.np.empty(0)

        adjoint = 0

        state = self.cast(state)
        gate = self.cast(gate)

        assert state.dtype == gate.dtype
        data_type, compute_type = self.get_cuda_type(state.dtype)

        if kernel == "apply_swap":
            nBitSwaps = 1
            bitSwaps = [(target1, target2)]
            maskLen = ncontrols
            maskBitString = self.np.ones(ncontrols)
            maskOrdering = controls

            self.cusv.swap_index_bits(
                self.handle,
                state.data.ptr,
                data_type,
                nqubits,
                bitSwaps,
                nBitSwaps,
                maskBitString,
                maskOrdering,
                maskLen,
            )
            return state

        if isinstance(gate, self.cp.ndarray):
            gate_ptr = gate.data.ptr
        elif isinstance(gate, self.np.ndarray):
            gate_ptr = gate.ctypes.data
        else:
            raise ValueError

        workspaceSize = self.cusv.apply_matrix_get_workspace_size(
            self.handle,
            data_type,
            nqubits,
            gate_ptr,
            data_type,
            self.cusv.MatrixLayout.ROW,
            adjoint,
            ntarget,
            ncontrols,
            compute_type,
        )

        # check the size of external workspace
        if workspaceSize > 0:
            workspace = self.cp.cuda.memory.alloc(workspaceSize)
            workspace_ptr = workspace.ptr
        else:
            workspace_ptr = 0

        self.cusv.apply_matrix(
            self.handle,
            state.data.ptr,
            data_type,
            nqubits,
            gate_ptr,
            data_type,
            self.cusv.MatrixLayout.ROW,
            adjoint,
            target.ctypes.data,
            ntarget,
            controls.ctypes.data,
            0,
            ncontrols,
            compute_type,
            workspace_ptr,
            workspaceSize,
        )

        return state

    def multi_qubit_base(self, state, nqubits, targets, gate, qubits=None):
        state = self.cast(state)
        ntarget = len(targets)
        if qubits is None:
            qubits = sorted(nqubits - q - 1 for q in targets)
        else:
            qubits = self.to_numpy(qubits)
        target = [nqubits - q - 1 for q in targets]
        target = self.np.asarray(target[::-1], dtype=self.np.int32)
        controls = self.np.asarray(
            [i for i in qubits if i not in target], dtype=self.np.int32
        )
        ncontrols = len(controls)
        adjoint = 0
        gate = self.cast(gate)
        assert state.dtype == gate.dtype
        data_type, compute_type = self.get_cuda_type(state.dtype)

        if isinstance(gate, self.cp.ndarray):
            gate_ptr = gate.data.ptr
        elif isinstance(gate, self.np.ndarray):
            gate_ptr = gate.ctypes.data
        else:
            raise ValueError

        workspaceSize = self.cusv.apply_matrix_get_workspace_size(
            self.handle,
            data_type,
            nqubits,
            gate_ptr,
            data_type,
            self.cusv.MatrixLayout.ROW,
            adjoint,
            ntarget,
            ncontrols,
            compute_type,
        )

        # check the size of external workspace
        if workspaceSize > 0:
            workspace = self.cp.cuda.memory.alloc(workspaceSize)
            workspace_ptr = workspace.ptr
        else:
            workspace_ptr = 0

        self.cusv.apply_matrix(
            self.handle,
            state.data.ptr,
            data_type,
            nqubits,
            gate_ptr,
            data_type,
            self.cusv.MatrixLayout.ROW,
            adjoint,
            target.ctypes.data,
            ntarget,
            controls.ctypes.data,
            0,
            ncontrols,
            compute_type,
            workspace_ptr,
            workspaceSize,
        )

        return state

    def collapse_state(self, state, qubits, shot, nqubits, normalize=True):
        state = self.cast(state)
        results = bin(int(shot)).replace("0b", "")
        results = list(map(int, "0" * (len(qubits) - len(results)) + results))[::-1]
        ntarget = 1
        qubits = self.np.asarray(
            [nqubits - q - 1 for q in reversed(qubits)], dtype="int32"
        )
        data_type, compute_type = self.get_cuda_type(state.dtype)

        for i in range(len(results)):
            self.cusv.collapse_on_z_basis(
                self.handle,
                state.data.ptr,
                data_type,
                nqubits,
                results[i],
                [qubits[i]],
                ntarget,
                1,
            )

        if normalize:
            norm = self.cp.sqrt(self.cp.sum(self.cp.square(self.cp.abs(state))))
            state = state / norm

        return state


class MultiGpuOps:  # pragma: no cover
    # CI does not have GPUs

    def __init__(self, backend, cpu_backend, circuit):
        self.backend = backend
        self.circuit = circuit
        self.cpu_ops = cpu_backend.ops

    def transpose_state(self, pieces, state, nqubits, order):
        original_shape = state.shape
        state = state.ravel()
        # always fall back to numba CPU backend because for ops not implemented on GPU
        state = self.cpu_ops.transpose_state(tuple(pieces), state, nqubits, order)
        return np.reshape(state, original_shape)

    def to_pieces(self, state):
        nqubits = self.circuit.nqubits
        qubits = self.circuit.queues.qubits
        shape = (self.circuit.ndevices, 2**self.circuit.nlocal)
        state = np.reshape(self.backend.to_numpy(state), shape)
        pieces = [state[i] for i in range(self.circuit.ndevices)]
        new_tensor = np.zeros(shape, dtype=state.dtype)
        new_tensor = self.transpose_state(
            pieces, new_tensor, nqubits, qubits.transpose_order
        )
        for i in range(self.circuit.ndevices):
            pieces[i] = new_tensor[i]
        return pieces

    def to_tensor(self, pieces):
        nqubits = self.circuit.nqubits
        qubits = self.circuit.queues.qubits
        if qubits.list == list(range(self.circuit.nglobal)):
            tensor = np.concatenate([x[np.newaxis] for x in pieces], axis=0)
            tensor = np.reshape(tensor, (2**nqubits,))
        elif qubits.list == list(range(self.circuit.nlocal, self.circuit.nqubits)):
            tensor = np.concatenate([x[:, np.newaxis] for x in pieces], axis=1)
            tensor = np.reshape(tensor, (2**nqubits,))
        else:  # fall back to the transpose op
            tensor = np.zeros(2**nqubits, dtype=self.backend.dtype)
            tensor = self.transpose_state(
                pieces, tensor, nqubits, qubits.reverse_transpose_order
            )
        return tensor

    def apply_gates(self, pieces, queues, ids, device):
        """Method that is parallelized using ``joblib``."""
        for i in ids:
            device_id = int(device.split(":")[-1]) % self.backend.ngpus
            with self.backend.cp.cuda.Device(device_id):
                piece = self.backend.cast(pieces[i])
                for gate in queues[i]:
                    piece = self.backend.apply_gate(gate, piece, self.circuit.nlocal)
            pieces[i] = self.backend.to_numpy(piece)
            del piece

    def apply_special_gate(self, pieces, gate):
        """Executes special gates on CPU.

        Currently special gates are ``Flatten`` or ``CallbackGate``.
        This method calculates the full state vector because special gates
        are not implemented for state pieces.
        """
        from qibo.gates import CallbackGate

        # Reverse all global SWAPs that happened so far
        pieces = self.revert_swaps(pieces, reversed(gate.swap_reset))
        state = self.to_tensor(pieces)
        if isinstance(gate, CallbackGate):
            gate.apply(self.backend, state, self.circuit.nqubits)
        else:
            state = gate.apply(self.backend, state, self.circuit.nqubits)
            pieces = self.to_pieces(state)
        # Redo all global SWAPs that happened so far
        pieces = self.revert_swaps(pieces, gate.swap_reset)
        return pieces

    def swap(self, pieces, global_qubit, local_qubit):
        m = self.circuit.queues.qubits.reduced_global.get(global_qubit)
        m = self.circuit.nglobal - m - 1
        t = 1 << m
        for g in range(self.circuit.ndevices // 2):
            i = ((g >> m) << (m + 1)) + (g & (t - 1))
            local_eff = self.circuit.queues.qubits.reduced_local.get(local_qubit)
            self.cpu_ops.swap_pieces(
                pieces[i], pieces[i + t], local_eff, self.circuit.nlocal
            )
        return pieces

    def revert_swaps(self, pieces, swap_pairs):
        for q1, q2 in swap_pairs:
            if q1 not in self.circuit.queues.qubits.set:
                q1, q2 = q2, q1
            pieces = self.swap(pieces, q1, q2)
        return pieces



import pytest

from qibojit.backends import CupyBackend, CuQuantumBackend, NumbaBackend

BACKENDS = {"numba": NumbaBackend, "cupy": CupyBackend, "cuquantum": CuQuantumBackend}

# ignore backends that are not available in the current testing environment
AVAILABLE_BACKENDS = []
for backend_name in BACKENDS.keys():
    try:
        BACKENDS.get(backend_name)()
        AVAILABLE_BACKENDS.append(backend_name)
    except (ModuleNotFoundError, ImportError):
        pass


@pytest.fixture
def backend(backend_name):
    yield BACKENDS.get(backend_name)()


def pytest_generate_tests(metafunc):
    if "backend_name" in metafunc.fixturenames:
        metafunc.parametrize("backend_name", AVAILABLE_BACKENDS)
    if "dtype" in metafunc.fixturenames:
        metafunc.parametrize("dtype", ["complex128", "complex64"])



import numpy as np
from numba import njit, prange


@njit("int64(int64, int32[:])", cache=True)
def multicontrol_index(g, qubits):
    i = 0
    i += g
    for n in qubits:
        k = 1 << n
        i = ((i >> n) << (n + 1)) + (i & (k - 1)) + k
    return i


@njit(
    [
        "complex64[:](complex64[:], complex64[:,:], int64, int64)",
        "complex128[:](complex128[:], complex128[:,:], int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def apply_gate_kernel(state, gate, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i1 = ((g >> m) << (m + 1)) + (g & (tk - 1))
        i2 = i1 + tk
        state[i1], state[i2] = (
            gate[0, 0] * state[i1] + gate[0, 1] * state[i2],
            gate[1, 0] * state[i1] + gate[1, 1] * state[i2],
        )
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64[:,:], int32[:], int64, int64)",
        "complex128[:](complex128[:], complex128[:,:], int32[:], int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def multicontrol_apply_gate_kernel(state, gate, qubits, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = multicontrol_index(g, qubits)
        i1, i2 = i - tk, i
        state[i1], state[i2] = (
            gate[0, 0] * state[i1] + gate[0, 1] * state[i2],
            gate[1, 0] * state[i1] + gate[1, 1] * state[i2],
        )
    return state


@njit(
    [
        "complex64[:](complex64[:], optional(complex64[:,:]), int64, int64)",
        "complex128[:](complex128[:], optional(complex128[:,:]), int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def apply_x_kernel(state, gate, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i1 = ((g >> m) << (m + 1)) + (g & (tk - 1))
        i2 = i1 + tk
        state[i1], state[i2] = state[i2], state[i1]
    return state


@njit(
    [
        "complex64[:](complex64[:], optional(complex64[:,:]), int32[:], int64, int64)",
        "complex128[:](complex128[:], optional(complex128[:,:]), int32[:], int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def multicontrol_apply_x_kernel(state, gate, qubits, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = multicontrol_index(g, qubits)
        i1, i2 = i - tk, i
        state[i1], state[i2] = state[i2], state[i1]
    return state


@njit(
    [
        "complex64[:](complex64[:], optional(complex64[:,:]), int64, int64)",
        "complex128[:](complex128[:], optional(complex128[:,:]), int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def apply_y_kernel(state, gate, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i1 = ((g >> m) << (m + 1)) + (g & (tk - 1))
        i2 = i1 + tk
        state[i1], state[i2] = -1j * state[i2], 1j * state[i1]
    return state


@njit(
    [
        "complex64[:](complex64[:], optional(complex64[:,:]), int32[:], int64, int64)",
        "complex128[:](complex128[:], optional(complex128[:,:]), int32[:], int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def multicontrol_apply_y_kernel(state, gate, qubits, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = multicontrol_index(g, qubits)
        i1, i2 = i - tk, i
        state[i1], state[i2] = -1j * state[i2], 1j * state[i1]
    return state


@njit(
    [
        "complex64[:](complex64[:], optional(complex64[:,:]), int64, int64)",
        "complex128[:](complex128[:], optional(complex128[:,:]), int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def apply_z_kernel(state, gate, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = ((g >> m) << (m + 1)) + (g & (tk - 1))
        state[i + tk] *= -1
    return state


@njit(
    [
        "complex64[:](complex64[:], optional(complex64[:,:]), int32[:], int64, int64)",
        "complex128[:](complex128[:], optional(complex128[:,:]), int32[:], int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def multicontrol_apply_z_kernel(state, gate, qubits, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = multicontrol_index(g, qubits)
        state[i] *= -1
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64, int64, int64)",
        "complex128[:](complex128[:], complex128, int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def apply_z_pow_kernel(state, gate, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = ((g >> m) << (m + 1)) + (g & (tk - 1))
        state[i + tk] = gate * state[i + tk]
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64, int32[:], int64, int64)",
        "complex128[:](complex128[:], complex128, int32[:], int64, int64)",
    ],
    parallel=True,
    cache=True,
)
def multicontrol_apply_z_pow_kernel(state, gate, qubits, nstates, m):
    tk = 1 << m
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = multicontrol_index(g, qubits)
        state[i] = gate * state[i]
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64[:,:], int64, int64, int64, boolean)",
        "complex128[:](complex128[:], complex128[:,:], int64, int64, int64, boolean)",
    ],
    parallel=True,
    cache=True,
)
def apply_two_qubit_gate_kernel(state, gate, nstates, m1, m2, swap_targets=False):
    tk1, tk2 = 1 << m1, 1 << m2
    uk1, uk2 = tk1, tk2
    if swap_targets:
        uk1, uk2 = uk2, uk1
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = ((g >> m1) << (m1 + 1)) + (g & (tk1 - 1))
        i = ((i >> m2) << (m2 + 1)) + (i & (tk2 - 1))
        i1, i2 = i + uk1, i + uk2
        i3 = i + tk1 + tk2
        buffer0, buffer1, buffer2 = state[i], state[i1], state[i2]
        state[i] = (
            gate[0, 0] * state[i]
            + gate[0, 1] * state[i1]
            + gate[0, 2] * state[i2]
            + gate[0, 3] * state[i3]
        )
        state[i1] = (
            gate[1, 0] * buffer0
            + gate[1, 1] * state[i1]
            + gate[1, 2] * state[i2]
            + gate[1, 3] * state[i3]
        )
        state[i2] = (
            gate[2, 0] * buffer0
            + gate[2, 1] * buffer1
            + gate[2, 2] * state[i2]
            + gate[2, 3] * state[i3]
        )
        state[i3] = (
            gate[3, 0] * buffer0
            + gate[3, 1] * buffer1
            + gate[3, 2] * buffer2
            + gate[3, 3] * state[i3]
        )
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64[:,:], int32[:], int64, int64, int64, boolean)",
        "complex128[:](complex128[:], complex128[:,:], int32[:], int64, int64, int64, boolean)",
    ],
    parallel=True,
    cache=True,
)
def multicontrol_apply_two_qubit_gate_kernel(
    state, gate, qubits, nstates, m1, m2, swap_targets=False
):
    tk1, tk2 = 1 << m1, 1 << m2
    uk1, uk2 = tk1, tk2
    if swap_targets:
        uk1, uk2 = uk2, uk1
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = multicontrol_index(g, qubits)
        i1, i2 = i - uk2, i - uk1
        i0 = i1 - uk1
        buffer0, buffer1, buffer2 = state[i0], state[i1], state[i2]
        state[i0] = (
            gate[0, 0] * state[i0]
            + gate[0, 1] * state[i1]
            + gate[0, 2] * state[i2]
            + gate[0, 3] * state[i]
        )
        state[i1] = (
            gate[1, 0] * buffer0
            + gate[1, 1] * state[i1]
            + gate[1, 2] * state[i2]
            + gate[1, 3] * state[i]
        )
        state[i2] = (
            gate[2, 0] * buffer0
            + gate[2, 1] * buffer1
            + gate[2, 2] * state[i2]
            + gate[2, 3] * state[i]
        )
        state[i] = (
            gate[3, 0] * buffer0
            + gate[3, 1] * buffer1
            + gate[3, 2] * buffer2
            + gate[3, 3] * state[i]
        )
    return state


@njit(
    [
        "complex64[:](complex64[:], optional(complex64[:,:]), int64, int64, int64, boolean)",
        "complex128[:](complex128[:], optional(complex128[:,:]), int64, int64, int64, boolean)",
    ],
    parallel=True,
    cache=True,
)
def apply_swap_kernel(state, gate, nstates, m1, m2, swap_targets=False):
    tk1, tk2 = 1 << m1, 1 << m2
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = ((g >> m1) << (m1 + 1)) + (g & (tk1 - 1))
        i = ((i >> m2) << (m2 + 1)) + (i & (tk2 - 1))
        i1, i2 = i + tk1, i + tk2
        state[i1], state[i2] = state[i2], state[i1]
    return state


@njit(
    [
        "complex64[:](complex64[:], optional(complex64[:,:]), int32[:], int64, int64, int64, boolean)",
        "complex128[:](complex128[:], optional(complex128[:,:]), int32[:], int64, int64, int64, boolean)",
    ],
    parallel=True,
    cache=True,
)
def multicontrol_apply_swap_kernel(
    state, gate, qubits, nstates, m1, m2, swap_targets=False
):
    tk1, tk2 = 1 << m1, 1 << m2
    uk1, uk2 = tk1, tk2
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = multicontrol_index(g, qubits)
        i1, i2 = i - tk2, i - tk1
        state[i1], state[i2] = state[i2], state[i1]
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64[:], int64, int64, int64, boolean)",
        "complex128[:](complex128[:], complex128[:], int64, int64, int64, boolean)",
    ],
    parallel=True,
    cache=True,
)
def apply_fsim_kernel(state, gate, nstates, m1, m2, swap_targets=False):
    tk1, tk2 = 1 << m1, 1 << m2
    uk1, uk2 = tk1, tk2
    if swap_targets:
        uk1, uk2 = uk2, uk1
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = ((g >> m1) << (m1 + 1)) + (g & (tk1 - 1))
        i = ((i >> m2) << (m2 + 1)) + (i & (tk2 - 1))
        i1, i2 = i + uk1, i + uk2
        i3 = i + tk1 + tk2
        state[i1], state[i2] = (
            gate[0] * state[i1] + gate[1] * state[i2],
            gate[2] * state[i1] + gate[3] * state[i2],
        )
        state[i3] *= gate[4]
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64[:], int32[:], int64, int64, int64, boolean)",
        "complex128[:](complex128[:], complex128[:], int32[:], int64, int64, int64, boolean)",
    ],
    parallel=True,
    cache=True,
)
def multicontrol_apply_fsim_kernel(
    state, gate, qubits, nstates, m1, m2, swap_targets=False
):
    tk1, tk2 = 1 << m1, 1 << m2
    uk1, uk2 = tk1, tk2
    if swap_targets:
        uk1, uk2 = uk2, uk1
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        i = multicontrol_index(g, qubits)
        i1, i2 = i - uk2, i - uk1
        state[i1], state[i2] = (
            gate[0] * state[i1] + gate[1] * state[i2],
            gate[2] * state[i1] + gate[3] * state[i2],
        )
        state[i] *= gate[4]
    return state


@njit("int64(int64, int64[:])", cache=True)
def multitarget_index(i, targets):
    t = 0
    for u, v in enumerate(targets):
        t += ((i >> u) & 1) * v
    return t


@njit(
    [
        "complex64[:](complex64[:], complex64[:,:], int32[:], int64, int64[:])",
        "complex128[:](complex128[:], complex128[:,:], int32[:], int64, int64[:])",
    ],
    parallel=True,
    cache=True,
)
def apply_three_qubit_gate_kernel(state, gate, qubits, nstates, targets):
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        ig = multicontrol_index(g, qubits)
        buffer0 = state[ig - targets[0] - targets[1] - targets[2]]
        buffer1 = state[ig - targets[1] - targets[2]]
        buffer2 = state[ig - targets[0] - targets[2]]
        buffer3 = state[ig - targets[2]]
        buffer4 = state[ig - targets[0] - targets[1]]
        buffer5 = state[ig - targets[1]]
        buffer6 = state[ig - targets[0]]
        buffer7 = state[ig]
        for i in range(8):
            t = ig - multitarget_index(7 - i, targets)
            state[t] = (
                gate[i, 0] * buffer0
                + gate[i, 1] * buffer1
                + gate[i, 2] * buffer2
                + gate[i, 3] * buffer3
                + gate[i, 4] * buffer4
                + gate[i, 5] * buffer5
                + gate[i, 6] * buffer6
                + gate[i, 7] * buffer7
            )
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64[:,:], int32[:], int64, int64[:])",
        "complex128[:](complex128[:], complex128[:,:], int32[:], int64, int64[:])",
    ],
    parallel=True,
    cache=True,
)
def apply_four_qubit_gate_kernel(state, gate, qubits, nstates, targets):
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        ig = multicontrol_index(g, qubits)
        buffer0 = state[ig - targets[0] - targets[1] - targets[2] - targets[3]]
        buffer1 = state[ig - targets[1] - targets[2] - targets[3]]
        buffer2 = state[ig - targets[0] - targets[2] - targets[3]]
        buffer3 = state[ig - targets[2] - targets[3]]
        buffer4 = state[ig - targets[0] - targets[1] - targets[3]]
        buffer5 = state[ig - targets[1] - targets[3]]
        buffer6 = state[ig - targets[0] - targets[3]]
        buffer7 = state[ig - targets[3]]
        buffer8 = state[ig - targets[0] - targets[1] - targets[2]]
        buffer9 = state[ig - targets[1] - targets[2]]
        buffer10 = state[ig - targets[0] - targets[2]]
        buffer11 = state[ig - targets[2]]
        buffer12 = state[ig - targets[0] - targets[1]]
        buffer13 = state[ig - targets[1]]
        buffer14 = state[ig - targets[0]]
        buffer15 = state[ig]
        for i in range(16):
            t = ig - multitarget_index(15 - i, targets)
            state[t] = (
                gate[i, 0] * buffer0
                + gate[i, 1] * buffer1
                + gate[i, 2] * buffer2
                + gate[i, 3] * buffer3
                + gate[i, 4] * buffer4
                + gate[i, 5] * buffer5
                + gate[i, 6] * buffer6
                + gate[i, 7] * buffer7
                + gate[i, 8] * buffer8
                + gate[i, 9] * buffer9
                + gate[i, 10] * buffer10
                + gate[i, 11] * buffer11
                + gate[i, 12] * buffer12
                + gate[i, 13] * buffer13
                + gate[i, 14] * buffer14
                + gate[i, 15] * buffer15
            )
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64[:,:], int32[:], int64, int64[:])",
        "complex128[:](complex128[:], complex128[:,:], int32[:], int64, int64[:])",
    ],
    parallel=True,
    cache=True,
)
def apply_five_qubit_gate_kernel(state, gate, qubits, nstates, targets):
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        ig = multicontrol_index(g, qubits)
        buffer0 = state[
            ig - targets[0] - targets[1] - targets[2] - targets[3] - targets[4]
        ]
        buffer1 = state[ig - targets[1] - targets[2] - targets[3] - targets[4]]
        buffer2 = state[ig - targets[0] - targets[2] - targets[3] - targets[4]]
        buffer3 = state[ig - targets[2] - targets[3] - targets[4]]
        buffer4 = state[ig - targets[0] - targets[1] - targets[3] - targets[4]]
        buffer5 = state[ig - targets[1] - targets[3] - targets[4]]
        buffer6 = state[ig - targets[0] - targets[3] - targets[4]]
        buffer7 = state[ig - targets[3] - targets[4]]
        buffer8 = state[ig - targets[0] - targets[1] - targets[2] - targets[4]]
        buffer9 = state[ig - targets[1] - targets[2] - targets[4]]
        buffer10 = state[ig - targets[0] - targets[2] - targets[4]]
        buffer11 = state[ig - targets[2] - targets[4]]
        buffer12 = state[ig - targets[0] - targets[1] - targets[4]]
        buffer13 = state[ig - targets[1] - targets[4]]
        buffer14 = state[ig - targets[0] - targets[4]]
        buffer15 = state[ig - targets[4]]
        buffer16 = state[ig - targets[0] - targets[1] - targets[2] - targets[3]]
        buffer17 = state[ig - targets[1] - targets[2] - targets[3]]
        buffer18 = state[ig - targets[0] - targets[2] - targets[3]]
        buffer19 = state[ig - targets[2] - targets[3]]
        buffer20 = state[ig - targets[0] - targets[1] - targets[3]]
        buffer21 = state[ig - targets[1] - targets[3]]
        buffer22 = state[ig - targets[0] - targets[3]]
        buffer23 = state[ig - targets[3]]
        buffer24 = state[ig - targets[0] - targets[1] - targets[2]]
        buffer25 = state[ig - targets[1] - targets[2]]
        buffer26 = state[ig - targets[0] - targets[2]]
        buffer27 = state[ig - targets[2]]
        buffer28 = state[ig - targets[0] - targets[1]]
        buffer29 = state[ig - targets[1]]
        buffer30 = state[ig - targets[0]]
        buffer31 = state[ig]
        for i in range(32):
            t = ig - multitarget_index(31 - i, targets)
            state[t] = (
                gate[i, 0] * buffer0
                + gate[i, 1] * buffer1
                + gate[i, 2] * buffer2
                + gate[i, 3] * buffer3
                + gate[i, 4] * buffer4
                + gate[i, 5] * buffer5
                + gate[i, 6] * buffer6
                + gate[i, 7] * buffer7
                + gate[i, 8] * buffer8
                + gate[i, 9] * buffer9
                + gate[i, 10] * buffer10
                + gate[i, 11] * buffer11
                + gate[i, 12] * buffer12
                + gate[i, 13] * buffer13
                + gate[i, 14] * buffer14
                + gate[i, 15] * buffer15
                + gate[i, 16] * buffer16
                + gate[i, 17] * buffer17
                + gate[i, 18] * buffer18
                + gate[i, 19] * buffer19
                + gate[i, 20] * buffer20
                + gate[i, 21] * buffer21
                + gate[i, 22] * buffer22
                + gate[i, 23] * buffer23
                + gate[i, 24] * buffer24
                + gate[i, 25] * buffer25
                + gate[i, 26] * buffer26
                + gate[i, 27] * buffer27
                + gate[i, 28] * buffer28
                + gate[i, 29] * buffer29
                + gate[i, 30] * buffer30
                + gate[i, 31] * buffer31
            )
    return state


@njit(
    [
        "complex64[:](complex64[:], complex64[:,::1], int32[:], int64, int64[:])",
        "complex128[:](complex128[:], complex128[:,::1], int32[:], int64, int64[:])",
    ],
    parallel=True,
    cache=True,
)
def apply_multi_qubit_gate_kernel(state, gate, qubits, nstates, targets):
    nsubstates = 1 << len(targets)
    for g in prange(nstates):  # pylint: disable=not-an-iterable
        ig = multicontrol_index(g, qubits)
        buffer = np.empty(nsubstates, dtype=state.dtype)
        for i in range(nsubstates):
            t = ig - multitarget_index(nsubstates - i - 1, targets)
            buffer[i] = state[t]
        for i in range(nsubstates):
            t = ig - multitarget_index(nsubstates - i - 1, targets)
            state[t] = np.dot(gate[i], buffer)
    return state



import numpy as np


def qubits_tensor(nqubits, targets, controls=[]):
    qubits = [nqubits - q - 1 for q in targets]
    qubits.extend(nqubits - q - 1 for q in controls)
    return np.array(sorted(qubits), dtype="int32")


def random_complex(shape, dtype="complex128"):
    x = np.random.random(shape) + 1j * np.random.random(shape)
    return x.astype(dtype)


def set_precision(dtype, *backends):
    if dtype == "complex64":
        precision = "single"
    else:
        precision = "double"
    for backend in backends:
        backend.set_precision(precision)



from typing import Union

from qibojit.backends.cpu import NumbaBackend
from qibojit.backends.gpu import CupyBackend, CuQuantumBackend

QibojitBackend = Union[NumbaBackend, CupyBackend, CuQuantumBackend]

PLATFORMS = ("numba", "cupy", "cuquantum")


class MetaBackend:
    """Meta-backend class which takes care of loading the qibojit backends."""

    @staticmethod
    def load(platform: str = None) -> QibojitBackend:
        """Loads the backend.

        Args:
            platform (str): Name of the backend to load: either `numba`, `cupy` or `cuquantum`.
        Returns:
            qibo.backends.abstract.Backend: The loaded backend.
        """

        if platform == "numba":
            return NumbaBackend()
        elif platform == "cupy":
            return CupyBackend()
        elif platform == "cuquantum":
            return CuQuantumBackend()
        else:  # pragma: no cover
            try:
                return CupyBackend()
            except (ModuleNotFoundError, ImportError):
                return NumbaBackend()

    def list_available(self) -> dict:
        """Lists all the available qibojit backends."""
        available_backends = {}
        for platform in PLATFORMS:
            try:
                MetaBackend.load(platform=platform)
                available = True
            except:
                available = False
            available_backends[platform] = available
        return available_backends



"""Set of custom Numba operations for the Clifford backend."""

import numpy as np
from numba import njit, prange, uint64

PARALLEL = False


@njit("Tuple((u1[:], u1[:,:], u1[:,:]))(u1[:,:], u8)", parallel=PARALLEL, cache=True)
def _get_rxz(symplectic_matrix, nqubits):
    return (
        symplectic_matrix[:, -1],
        symplectic_matrix[:, :nqubits],
        symplectic_matrix[:, nqubits:-1],
    )


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def H(symplectic_matrix, q, nqubits):
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = r[i] ^ (x[i, q] & z[i, q])
        tmp = symplectic_matrix[i, q]
        symplectic_matrix[i, q] = symplectic_matrix[i, nqubits + q]
        symplectic_matrix[i, nqubits + q] = tmp
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8, u8)", parallel=PARALLEL, cache=True)
def CNOT(symplectic_matrix, control_q, target_q, nqubits):
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        r[i] = r[i] ^ (x[i, control_q] & z[i, target_q]) & (
            x[i, target_q] ^ ~z[i, control_q]
        )
        x[i, target_q] = x[i, target_q] ^ x[i, control_q]
        z[i, control_q] = z[i, control_q] ^ z[i, target_q]
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8, u8)", parallel=PARALLEL, cache=True)
def CZ(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> H-CNOT-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = (
            r[i]
            ^ (x[i, target_q] & z[i, target_q])
            ^ (x[i, control_q] & x[i, target_q] & (z[i, target_q] ^ ~z[i, control_q]))
            ^ (x[i, target_q] & (z[i, target_q] ^ x[i, control_q]))
        )
        z_control_q = x[i, target_q] ^ z[i, control_q]
        z_target_q = z[i, target_q] ^ x[i, control_q]
        z[i, control_q] = z_control_q
        z[i, target_q] = z_target_q
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def S(symplectic_matrix, q, nqubits):
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = r[i] ^ (x[i, q] & z[i, q])
        symplectic_matrix[i, nqubits + q] = z[i, q] ^ x[i, q]
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def Z(symplectic_matrix, q, nqubits):
    """Decomposition --> S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = r[i] ^ (
            (x[i, q] & z[i, q]) ^ x[i, q] & (z[i, q] ^ x[i, q])
        )
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def X(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-S-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = (
            r[i] ^ (z[i, q] & (z[i, q] ^ x[i, q])) ^ (z[i, q] & x[i, q])
        )
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def Y(symplectic_matrix, q, nqubits):
    """Decomposition --> S-S-H-S-S-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = (
            r[i] ^ (z[i, q] & (z[i, q] ^ x[i, q])) ^ (x[i, q] & (z[i, q] ^ x[i, q]))
        )
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def SX(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = r[i] ^ (z[i, q] & (z[i, q] ^ x[i, q]))
        symplectic_matrix[i, q] = z[i, q] ^ x[i, q]
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def SDG(symplectic_matrix, q, nqubits):
    """Decomposition --> S-S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = r[i] ^ (x[i, q] & (z[i, q] ^ x[i, q]))
        symplectic_matrix[i, nqubits + q] = z[i, q] ^ x[i, q]
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def SXDG(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-S-S-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = r[i] ^ (z[i, q] & x[i, q])
        symplectic_matrix[i, q] = z[i, q] ^ x[i, q]
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def RY_pi(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = r[i] ^ (x[i, q] & (z[i, q] ^ x[i, q]))
        zq = symplectic_matrix[i, nqubits + q]
        symplectic_matrix[i, nqubits + q] = symplectic_matrix[i, q]
        symplectic_matrix[i, q] = zq
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8)", parallel=PARALLEL, cache=True)
def RY_3pi_2(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = r[i] ^ (z[i, q] & (z[i, q] ^ x[i, q]))
        zq = symplectic_matrix[i, nqubits + q]
        symplectic_matrix[i, nqubits + q] = symplectic_matrix[i, q]
        symplectic_matrix[i, q] = zq
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8, u8)", parallel=PARALLEL, cache=True)
def SWAP(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> CNOT-CNOT-CNOT"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = (
            r[i]
            ^ (x[i, control_q] & z[i, target_q] & (x[i, target_q] ^ ~z[i, control_q]))
            ^ (
                (x[i, target_q] ^ x[i, control_q])
                & (z[i, target_q] ^ z[i, control_q])
                & (z[i, target_q] ^ ~x[i, control_q])
            )
            ^ (
                x[i, target_q]
                & z[i, control_q]
                & (x[i, control_q] ^ x[i, target_q] ^ z[i, control_q] ^ ~z[i, target_q])
            )
        )
        x_cq = symplectic_matrix[i, control_q]
        x_tq = symplectic_matrix[i, target_q]
        z_cq = symplectic_matrix[i, nqubits + control_q]
        z_tq = symplectic_matrix[i, nqubits + target_q]
        symplectic_matrix[i, control_q] = x_tq
        symplectic_matrix[i, target_q] = x_cq
        symplectic_matrix[i, nqubits + control_q] = z_tq
        symplectic_matrix[i, nqubits + target_q] = z_cq
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8, u8)", parallel=PARALLEL, cache=True)
def iSWAP(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> H-CNOT-CNOT-H-S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = (
            r[i]
            ^ (x[i, target_q] & z[i, target_q])
            ^ (x[i, control_q] & z[i, control_q])
            ^ (x[i, control_q] & (z[i, control_q] ^ x[i, control_q]))
            ^ (
                (z[i, control_q] ^ x[i, control_q])
                & (z[i, target_q] ^ x[i, target_q])
                & (x[i, target_q] ^ ~x[i, control_q])
            )
            ^ (
                (x[i, target_q] ^ z[i, control_q] ^ x[i, control_q])
                & (x[i, target_q] ^ z[i, target_q] ^ x[i, control_q])
                & (x[i, target_q] ^ z[i, target_q] ^ x[i, control_q] ^ ~z[i, control_q])
            )
            ^ (x[i, control_q] & (x[i, target_q] ^ x[i, control_q] ^ z[i, control_q]))
        )
        z_control_q = x[i, target_q] ^ z[i, target_q] ^ x[i, control_q]
        z_target_q = x[i, target_q] ^ z[i, control_q] ^ x[i, control_q]
        symplectic_matrix[i, nqubits + control_q] = z_control_q
        symplectic_matrix[i, nqubits + target_q] = z_target_q
        tmp = symplectic_matrix[i, control_q]
        symplectic_matrix[i, control_q] = symplectic_matrix[i, target_q]
        symplectic_matrix[i, target_q] = tmp
    return symplectic_matrix


@njit("u1[:,:](u1[:,:], u8, u8, u8)", parallel=PARALLEL, cache=True)
def CY(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> S-CNOT-SDG"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)

    for i in prange(symplectic_matrix.shape[0]):  # pylint: disable=not-an-iterable
        symplectic_matrix[i, -1] = (
            r[i]
            ^ (x[i, target_q] & (z[i, target_q] ^ x[i, target_q]))
            ^ (
                x[i, control_q]
                & (x[i, target_q] ^ z[i, target_q])
                & (z[i, control_q] ^ ~x[i, target_q])
            )
            ^ ((x[i, target_q] ^ x[i, control_q]) & (z[i, target_q] ^ x[i, target_q]))
        )
        x_target_q = x[i, control_q] ^ x[i, target_q]
        z_control_q = z[i, control_q] ^ z[i, target_q] ^ x[i, target_q]
        z_target_q = z[i, target_q] ^ x[i, control_q]
        symplectic_matrix[i, target_q] = x_target_q
        symplectic_matrix[i, nqubits + control_q] = z_control_q
        symplectic_matrix[i, nqubits + target_q] = z_target_q
    return symplectic_matrix


@njit(
    [
        "u1[:,:](u1[:,:], u8[:], u8[:], u8, b1)",
        "u1[:,:](u1[:,:], u4[:], u4[:], u4, b1)",
    ],
    parallel=PARALLEL,
    cache=True,
    fastmath=True,
)
def _rowsum(symplectic_matrix, h, i, nqubits, determined=False):
    xi, zi = symplectic_matrix[i, :nqubits], symplectic_matrix[i, nqubits:-1]
    xh, zh = symplectic_matrix[h, :nqubits], symplectic_matrix[h, nqubits:-1]
    if determined:
        g_r = np.zeros(h.shape[0], dtype=np.uint8)
        g_xi_xh = xi.copy()
        g_zi_zh = xi.copy()
    for j in prange(len(h)):  # pylint: disable=not-an-iterable
        exp = (
            2 * (xi[j] * xh[j] * (zh[j] - zi[j]) + zi[j] * zh[j] * (xi[j] - xh[j]))
            - xi[j] * zh[j]
            + xh[j] * zi[j]
        )
        r = (
            2 * symplectic_matrix[h[j], -1]
            + 2 * symplectic_matrix[i[j], -1]
            + np.sum(exp)
        ) % 4 != 0
        xi_xh = xi[j] ^ xh[j]
        zi_zh = zi[j] ^ zh[j]
        if determined:  # for some reason xor reduction fails here
            g_r[j] = r  # thus, I cannot do g_r += r here
            g_xi_xh[j] = xi_xh
            g_zi_zh[j] = zi_zh
        else:
            symplectic_matrix[h[j], -1] = r
            symplectic_matrix[h[j], :nqubits] = xi_xh
            symplectic_matrix[h[j], nqubits:-1] = zi_zh
    if determined:
        for j in prange(len(g_r)):  # pylint: disable=not-an-iterable
            symplectic_matrix[h[0], -1] ^= g_r[j]
            symplectic_matrix[h[0], :nqubits] ^= g_xi_xh[j]
            symplectic_matrix[h[0], nqubits:-1] ^= g_zi_zh[j]
    return symplectic_matrix



import itertools

import numpy as np
import pytest
from qibo.quantum_info import random_density_matrix, random_statevector

from qibojit.tests.utils import qubits_tensor, random_complex, set_precision


@pytest.mark.parametrize("is_matrix", [False, True])
def test_zero_state(backend, dtype, is_matrix):
    set_precision(dtype, backend)
    if is_matrix:
        final_state = backend.zero_density_matrix(4)
        target_state = np.array([1] + [0] * 255, dtype=dtype)
        target_state = np.reshape(target_state, (16, 16))
    else:
        final_state = backend.zero_state(4)
        target_state = np.array([1] + [0] * 15, dtype=dtype)
    backend.assert_allclose(final_state, target_state)


@pytest.mark.parametrize("normalize", [False, True])
def test_identity_density_matrix(backend, dtype, normalize):
    set_precision(dtype, backend)
    norm = 16 if normalize else 1.0
    final_state = backend.identity_density_matrix(4, normalize=normalize)
    target_state = np.eye(16, dtype=dtype) / norm
    backend.assert_allclose(final_state, target_state)


@pytest.mark.parametrize("is_matrix", [False, True])
def test_plus_state(backend, dtype, is_matrix):
    set_precision(dtype, backend)
    if is_matrix:
        final_state = backend.plus_density_matrix(4)
        target_state = np.ones((16, 16), dtype=dtype) / 16
    else:
        final_state = backend.plus_state(4)
        target_state = np.ones(16, dtype=dtype) / 4
    backend.assert_allclose(final_state, target_state)


@pytest.mark.parametrize(
    "nqubits,targets,results",
    [
        (2, [0], [1]),
        (2, [1], [0]),
        (3, [1], [1]),
        (4, [1, 3], [1, 0]),
        (5, [1, 2, 4], [0, 1, 1]),
        (15, [4, 7], [0, 0]),
        (16, [8, 12, 15], [1, 0, 1]),
    ],
)
@pytest.mark.parametrize("normalize", [True, False])
def test_collapse_state(backend, nqubits, targets, results, normalize, dtype):
    atol = 1e-7 if dtype == "complex64" else 1e-14
    state = random_statevector(2**nqubits, backend=backend)
    slicer = nqubits * [slice(None)]
    for t, r in zip(targets, results):
        slicer[t] = r
    slicer = tuple(slicer)
    initial_state = np.reshape(np.copy(state), nqubits * (2,))
    target_state = np.zeros_like(initial_state)
    target_state[slicer] = initial_state[slicer]
    target_state = target_state.flatten()
    if normalize:
        norm = (np.abs(target_state) ** 2).sum()
        target_state = target_state / np.sqrt(norm)

    b2d = 2 ** np.arange(len(results) - 1, -1, -1)
    result = np.array([np.array(results).dot(b2d)])
    state = backend.collapse_state(state, sorted(targets), result, nqubits, normalize)
    backend.assert_allclose(state, target_state, atol=atol)


@pytest.mark.parametrize("density_matrix", [False, True])
def test_collapse_call(backend, density_matrix):
    from qibo import gates
    from qibo.backends import NumpyBackend

    tbackend = NumpyBackend()
    if density_matrix:
        state = random_density_matrix(2**3, backend=tbackend)
    else:
        state = random_statevector(2**3, backend=tbackend)

    gate = gates.M(0, 1, collapse=True)
    np.random.seed(123)
    if density_matrix:
        target_state = gate.apply_density_matrix(tbackend, np.copy(state), 3)
        np.random.seed(123)
        final_state = gate.apply_density_matrix(backend, np.copy(state), 3)
    else:
        target_state = gate.apply(tbackend, np.copy(state), 3)
        np.random.seed(123)
        final_state = gate.apply(backend, np.copy(state), 3)
    backend.assert_allclose(final_state, target_state)


def generate_transpose_qubits(nqubits):
    """Generates global qubits randomly."""
    qubits = np.arange(nqubits)
    np.random.shuffle(qubits)
    return qubits


CONFIG = ((n, generate_transpose_qubits(n)) for _ in range(5) for n in range(3, 11))


@pytest.mark.parametrize("nqubits,qubits", CONFIG)
@pytest.mark.parametrize("ndevices", [2, 4, 8])
def test_transpose_state(backend, nqubits, qubits, ndevices, dtype):
    if backend.platform != "numba":  # pragma: no cover
        pytest.skip(
            f"``transpose_state`` op is not available for {backend.platform} platform."
        )
    qubit_order = list(qubits)
    state = random_statevector(2**nqubits, backend=backend)
    state = backend.cast(state, dtype=dtype)
    state_tensor = np.reshape(state, nqubits * (2,))
    target_state = np.transpose(state_tensor, qubit_order).flatten()
    new_state = np.zeros_like(state)
    state = np.reshape(state, (ndevices, int(state.shape[0]) // ndevices))
    pieces = [state[i] for i in range(ndevices)]
    new_state = backend.ops.transpose_state(pieces, new_state, nqubits, qubit_order)
    backend.assert_allclose(new_state, target_state)


CONFIG = ((n, np.random.randint(1, n)) for _ in range(10) for n in range(4, 11))


@pytest.mark.parametrize("nqubits,local", CONFIG)
def test_swap_pieces_zero_global(backend, nqubits, local, dtype):
    if backend.platform != "numba":  # pragma: no cover
        pytest.skip(
            f"``swap_pieces`` op is not available for {backend.platform} platform."
        )

    from qibo import gates

    state = random_statevector(2**nqubits, backend=backend)
    state = backend.cast(state, dtype=dtype)
    target_state = np.copy(state)
    shape = (2, int(state.shape[0]) // 2)
    state = np.reshape(state, shape)

    set_precision(dtype, backend)
    gate = gates.SWAP(0, local)
    target_state = backend.apply_gate(gate, target_state, nqubits)
    target_state = np.reshape(backend.to_numpy(target_state), shape)
    piece0, piece1 = state[0], state[1]
    backend.ops.swap_pieces(piece0, piece1, local - 1, nqubits - 1)
    backend.assert_allclose(piece0, target_state[0])
    backend.assert_allclose(piece1, target_state[1])


CONFIG = (
    (n, np.random.randint(0, n), np.random.randint(0, n))
    for _ in range(10)
    for n in range(5, 11)
)


@pytest.mark.parametrize("nqubits,qlocal,qglobal", CONFIG)
def test_swap_pieces(backend, nqubits, qlocal, qglobal, dtype):
    if backend.platform != "numba":  # pragma: no cover
        pytest.skip(
            f"``swap_pieces`` op is not available for {backend.platform} platform."
        )

    from qibo import gates

    state = random_statevector(2**nqubits, backend=backend)
    state = backend.cast(state, dtype=dtype)
    target_state = np.copy(state)
    shape = (2, int(state.shape[0]) // 2)

    while qlocal == qglobal:
        qlocal = np.random.randint(0, nqubits)

    transpose_order = (
        [qglobal] + list(range(qglobal)) + list(range(qglobal + 1, nqubits))
    )

    gate = gates.SWAP(qglobal, qlocal)
    target_state = backend.apply_gate(gate, target_state, nqubits)
    target_state = backend.to_numpy(target_state)
    target_state = np.reshape(target_state, nqubits * (2,))
    target_state = np.transpose(target_state, transpose_order)
    target_state = np.reshape(target_state, shape)

    state = np.reshape(state, nqubits * (2,))
    state = np.transpose(state, transpose_order)
    state = np.reshape(state, shape)
    piece0, piece1 = state[0], state[1]
    new_global = qlocal - int(qglobal < qlocal)
    backend.ops.swap_pieces(piece0, piece1, new_global, nqubits - 1)
    backend.assert_allclose(piece0, target_state[0])
    backend.assert_allclose(piece1, target_state[1])


@pytest.mark.parametrize("realtype", ["float32", "float64"])
@pytest.mark.parametrize("inttype", ["int32", "int64"])
@pytest.mark.parametrize("nthreads", [None, 4])
def test_measure_frequencies(backend, realtype, inttype, nthreads):
    probs = np.ones(16, dtype=realtype) / 16
    frequencies = np.zeros(16, dtype=inttype)
    # if backend.name in ["cupy", "cuquantum"]:  # pragma: no cover
    # CI does not test for GPU
    #    with pytest.raises(NotImplementedError):
    #        frequencies = backend.measure_frequencies(frequencies, probs, nshots=1000,
    #                                                     nqubits=4, seed=1234,
    #                                                     nthreads=nthreads)
    # else:
    if nthreads is None:
        nthreads = backend.nthreads
    frequencies = backend.measure_frequencies_op(
        frequencies, probs, nshots=1000, nqubits=4, seed=1234, nthreads=nthreads
    )
    assert np.sum(frequencies) == 1000
    if nthreads == 4:
        target_frequencies = np.array(
            [72, 65, 63, 54, 57, 55, 67, 50, 53, 67, 69, 68, 64, 68, 66, 62],
            dtype=inttype,
        )
        backend.assert_allclose(frequencies, target_frequencies)


NONZERO = list(itertools.combinations(range(8), r=1))
NONZERO.extend(itertools.combinations(range(8), r=2))
NONZERO.extend(itertools.combinations(range(8), r=3))
NONZERO.extend(itertools.combinations(range(8), r=4))
NSHOTS = (len(NONZERO) // 2 + 1) * [1000, 200000]


@pytest.mark.parametrize("nonzero,nshots", zip(NONZERO, NSHOTS))
def test_measure_frequencies_sparse_probabilities(backend, nonzero, nshots):
    probs = np.zeros(8, dtype=np.float64)
    for i in nonzero:
        probs[i] = 1
    probs = probs / np.sum(probs)
    frequencies = backend.sample_frequencies(probs, nshots)
    assert sum(frequencies.values()) == nshots
    for i, freq in frequencies.items():
        assert freq != 0
